import collections
import logging

import six

from python_jsonschema_objects import util
from python_jsonschema_objects.validators import registry, ValidationError
from python_jsonschema_objects.util import lazy_format as fmt

logger = logging.getLogger(__name__)


class ArrayWrapper(collections.abc.MutableSequence):
    """A wrapper for array-like structures.

    This implements all of the array like behavior that one would want,
    with a dirty-tracking mechanism to avoid constant validation costs.
    """

    @property
    def strict(self):
        return getattr(self, "_strict_", False)

    def __len__(self):
        return len(self.data)

    def mark_or_revalidate(self):
        if self.strict:
            self.validate()
        else:
            self._dirty = True

    def __delitem__(self, index):
        self.data.pop(index)
        self.mark_or_revalidate()

    def insert(self, index, value):
        self.data.insert(index, value)
        self.mark_or_revalidate()

    def __setitem__(self, index, value):
        self.data[index] = value
        self.mark_or_revalidate()

    def __getitem__(self, idx):
        return self.typed_elems[idx]

    def __eq__(self, other):
        if isinstance(other, ArrayWrapper):
            return self.for_json() == other.for_json()
        else:
            return self.for_json() == other

    def __init__(self, ary):
        """Initialize a wrapper for the array

        Args:
            ary: (list-like, or ArrayWrapper)
        """

        """ Marks whether or not the underlying array has been modified """
        self._dirty = True

        """ Holds a typed copy of the array """
        self._typed = None

        if isinstance(ary, (list, tuple, collections.abc.Sequence)):
            self.data = ary
        else:
            raise TypeError("Invalid value given to array validator: {0}".format(ary))

        logger.debug(fmt("Initializing ArrayWrapper {} with {}", self, ary))

    @property
    def typed_elems(self):
        logger.debug(fmt("Accessing typed_elems of ArrayWrapper {} ", self))
        if self._typed is None or self._dirty is True:
            self.validate()

        return self._typed

    def __repr__(self):
        return "<%s=%s>" % (self.__class__.__name__, str(self.data))

    @classmethod
    def from_json(cls, jsonmsg):
        import json

        msg = json.loads(jsonmsg)
        obj = cls(msg)
        obj.validate()
        return obj

    def serialize(self):
        enc = util.ProtocolJSONEncoder()
        return enc.encode(self.typed_elems)

    def for_json(self):
        from python_jsonschema_objects import classbuilder

        out = []
        for item in self.typed_elems:
            if isinstance(
                item,
                (classbuilder.ProtocolBase, classbuilder.LiteralValue, ArrayWrapper),
            ):
                out.append(item.for_json())
            else:
                out.append(item)

        return out

    def validate(self):
        if self.strict or self._dirty:
            self.validate_items()
            self.validate_length()
            self.validate_uniqueness()
        return True

    def validate_uniqueness(self):

        if getattr(self, "uniqueItems", False) is True:
            testset = set(repr(item) for item in self.data)
            if len(testset) != len(self.data):
                raise ValidationError(
                    "{0} has duplicate elements, but uniqueness required".format(
                        self.data
                    )
                )

    def validate_length(self):

        if getattr(self, "minItems", None) is not None:
            if len(self.data) < self.minItems:
                raise ValidationError(
                    "{1} has too few elements. Wanted {0}.".format(
                        self.minItems, self.data
                    )
                )

        if getattr(self, "maxItems", None) is not None:
            if len(self.data) > self.maxItems:
                raise ValidationError(
                    "{1} has too many elements. Wanted {0}.".format(
                        self.maxItems, self.data
                    )
                )

    def validate_items(self):
        """Validates the items in the backing array, including
        performing type validation.

        Sets the _typed property and clears the dirty flag as a side effect

        Returns:
            The typed array
        """
        logger.debug(fmt("Validating {}", self))
        from python_jsonschema_objects import classbuilder

        if self.__itemtype__ is None:
            return

        type_checks = self.__itemtype__
        if not isinstance(type_checks, (tuple, list)):
            # we were given items = {'type': 'blah'} ; thus ensure the type for all data.
            type_checks = [type_checks] * len(self.data)
        elif len(type_checks) > len(self.data):
            raise ValidationError(
                "{1} does not have sufficient elements to validate against {0}".format(
                    self.__itemtype__, self.data
                )
            )

        typed_elems = []
        for elem, typ in zip(self.data, type_checks):
            if isinstance(typ, dict):
                for param, paramval in six.iteritems(typ):
                    validator = registry(param)
                    if validator is not None:
                        validator(paramval, elem, typ)
                typed_elems.append(elem)

            elif util.safe_issubclass(typ, classbuilder.LiteralValue):
                val = typ(elem)
                val.validate()
                typed_elems.append(val)
            elif util.safe_issubclass(typ, classbuilder.ProtocolBase):
                if not isinstance(elem, typ):
                    try:
                        if isinstance(
                            elem, (six.string_types, six.integer_types, float)
                        ):
                            val = typ(elem)
                        else:
                            val = typ(**util.coerce_for_expansion(elem))
                    except TypeError as e:
                        raise ValidationError(
                            "'{0}' is not a valid value for '{1}': {2}".format(
                                elem, typ, e
                            )
                        )
                else:
                    val = elem
                val.validate()
                typed_elems.append(val)

            elif util.safe_issubclass(typ, ArrayWrapper):
                val = typ(elem)
                val.validate()
                typed_elems.append(val)

            elif isinstance(typ, (classbuilder.TypeProxy, classbuilder.TypeRef)):
                try:
                    if isinstance(elem, (six.string_types, six.integer_types, float)):
                        val = typ(elem)
                    else:
                        val = typ(**util.coerce_for_expansion(elem))
                except TypeError as e:
                    raise ValidationError(
                        "'{0}' is not a valid value for '{1}': {2}".format(elem, typ, e)
                    )
                else:
                    val.validate()
                    typed_elems.append(val)

        self._dirty = False
        self._typed = typed_elems
        return typed_elems

    @staticmethod
    def create(name, item_constraint=None, **addl_constraints):
        """Create an array validator based on the passed in constraints.

        If item_constraint is a tuple, it is assumed that tuple validation
        is being performed. If it is a class or dictionary, list validation
        will be performed. Classes are assumed to be subclasses of ProtocolBase,
        while dictionaries are expected to be basic types ('string', 'number', ...).

        addl_constraints is expected to be key-value pairs of any of the other
        constraints permitted by JSON Schema v4.
        """
        logger.debug(
            fmt(
                "Constructing ArrayValidator with {} and {}",
                item_constraint,
                addl_constraints,
            )
        )
        from python_jsonschema_objects import classbuilder

        klassbuilder = addl_constraints.pop(
            "classbuilder", None
        )  # type: python_jsonschema_objects.classbuilder.ClassBuilder
        props = {}

        if item_constraint is not None:
            if isinstance(item_constraint, (tuple, list)):
                for i, elem in enumerate(item_constraint):
                    isdict = isinstance(elem, (dict,))
                    isklass = isinstance(elem, type) and util.safe_issubclass(
                        elem, (classbuilder.ProtocolBase, classbuilder.LiteralValue)
                    )

                    if not any([isdict, isklass]):
                        raise TypeError(
                            "Item constraint (position {0}) is not a schema".format(i)
                        )
            elif isinstance(
                item_constraint, (classbuilder.TypeProxy, classbuilder.TypeRef)
            ):
                pass
            elif util.safe_issubclass(item_constraint, ArrayWrapper):
                pass
            else:
                isdict = isinstance(item_constraint, (dict,))
                isklass = isinstance(item_constraint, type) and util.safe_issubclass(
                    item_constraint,
                    (classbuilder.ProtocolBase, classbuilder.LiteralValue),
                )

                if not any([isdict, isklass]):
                    raise TypeError("Item constraint is not a schema")

                if isdict and "$ref" in item_constraint:
                    if klassbuilder is None:
                        raise TypeError(
                            "Cannot resolve {0} without classbuilder".format(
                                item_constraint["$ref"]
                            )
                        )

                    item_constraint = klassbuilder.resolve_type(
                        item_constraint["$ref"], name
                    )

                elif isdict and item_constraint.get("type") == "array":
                    # We need to create a sub-array validator.
                    item_constraint = ArrayWrapper.create(
                        name + "#sub",
                        item_constraint=item_constraint["items"],
                        addl_constraints=item_constraint,
                    )
                elif isdict and "oneOf" in item_constraint:
                    # We need to create a TypeProxy validator
                    uri = "{0}_{1}".format(name, "<anonymous_list_type>")
                    type_array = klassbuilder.construct_objects(
                        item_constraint["oneOf"], uri
                    )

                    item_constraint = classbuilder.TypeProxy(type_array)

                elif isdict and item_constraint.get("type") == "object":
                    """ We need to create a ProtocolBase object for this anonymous definition"""
                    uri = "{0}_{1}".format(name, "<anonymous_list_type>")
                    item_constraint = klassbuilder.construct(uri, item_constraint)

        props["__itemtype__"] = item_constraint

        strict = addl_constraints.pop("strict", False)
        props["_strict_"] = strict
        props.update(addl_constraints)

        validator = type(str(name), (ArrayWrapper,), props)

        return validator



import pytest

import python_jsonschema_objects as pjo


@pytest.fixture
def classes():

    schema = {
        "title": "Config",
        "type": "object",
        "additionalProperties": {"$ref": "#/definitions/Parameter"},
        "definitions": {
            "Parameter": {
                "$id": "Parameter",
                "type": "object",
                "properties": {
                    "key": {"type": "string"},
                    "value": {"oneOf": [{"type": "string"}, {"type": "number"}]},
                },
            }
        },
    }

    return pjo.ObjectBuilder(schema).build_classes()


def test_pjo_objects_can_be_used_as_property_keys(classes):
    container = classes.Config()
    obj1 = classes.Parameter(key="volume", value=11)
    obj2 = classes.Parameter(key="compression", value="on")

    container[obj1.key] = obj1
    container[obj2.key] = obj2

    assert container.volume.value == 11

    assert sorted(container.keys()) == sorted(str(k) for k in [obj1.key, obj2.key])



import pytest  # noqa
import python_jsonschema_objects as pjo


def test_circular_references(markdown_examples):

    builder = pjo.ObjectBuilder(markdown_examples["Circular References"])
    klasses = builder.build_classes()
    a = klasses.A()
    b = klasses.B()
    a.message = "foo"
    b.author = "James Dean"
    a.reference = b
    a.reference = b

    assert a.reference == b
    assert b.oreference == None  # noqa
    assert a.message == "foo"

    serialized = a.serialize(sort_keys=True)
    assert (
        serialized == '{"message": "foo", "reference": {"author": "James Dean"}}'
    )  # noqa



import pytest
import python_jsonschema_objects as pjo


def test_114():
    schema = {
        "title": "Example",
        "type": "object",
        "properties": {
            "test_regression_114_anon_array": {
                "type": "array",
                "items": [
                    {
                        "oneOf": [
                            {
                                "type": "string",
                                "pattern": "^(?:(?:\\d|[1-9]\\d|1\\d\\d|2[0-4]\\d|25[0-5])\\.){3}(?:\\d|[1-9]\\d|1\\d\\d|2[0-4]\\d|25[0-5])$",
                            },
                            {
                                "type": "string",
                                "pattern": "^((([0-9A-Fa-f]{1,4}:){1,6}:)|(([0-9A-Fa-f]{1,4}:){7}))([0-9A-Fa-f]{1,4})$",
                            },
                        ]
                    }
                ],
            }
        },
    }

    builder = pjo.ObjectBuilder(schema)
    test = builder.build_classes()
    assert test



from python_jsonschema_objects import util
from python_jsonschema_objects import validators
import functools
import logging
import six
import operator


def MakeLiteral(name, typ, value, **properties):
    properties.update({"type": typ})
    klass = type(
        str(name),
        tuple((LiteralValue,)),
        {
            "__propinfo__": {
                "__literal__": properties,
                "__default__": properties.get("default"),
            }
        },
    )

    return klass(value)


@functools.total_ordering
class LiteralValue(object):
    """Docstring for LiteralValue """

    isLiteralClass = True

    def __init__(self, value, typ=None):
        """@todo: to be defined

        :value: @todo

        """
        if isinstance(value, LiteralValue):
            self._value = value._value
        else:
            self._value = value

        if self._value is None and self.default() is not None:
            self._value = self.default()

        self.validate()

    def as_dict(self):
        return self.for_json()

    def for_json(self):
        return self._value

    @classmethod
    def default(cls):
        return cls.__propinfo__.get("__default__")

    @classmethod
    def propinfo(cls, propname):
        if propname not in cls.__propinfo__:
            return {}
        return cls.__propinfo__[propname]

    def serialize(self):
        self.validate()
        enc = util.ProtocolJSONEncoder()
        return enc.encode(self)

    def __repr__(self):
        return "<Literal<%s> %s>" % (self._value.__class__.__name__, str(self._value))

    def __str__(self):
        if isinstance(self._value, six.string_types):
            return self._value
        return str(self._value)

    def validate(self):
        info = self.propinfo("__literal__")

        # TODO: this duplicates logic in validators.ArrayValidator.check_items; unify it.
        for param, paramval in sorted(
            six.iteritems(info), key=lambda x: x[0].lower() != "type"
        ):
            validator = validators.registry(param)
            if validator is not None:
                validator(paramval, self._value, info)

    def __eq__(self, other):
        if isinstance(other, LiteralValue):
            return self._value == other._value
        return self._value == other

    def __hash__(self):
        return hash(self._value)

    def __lt__(self, other):
        if isinstance(other, LiteralValue):
            return self._value < other._value
        return self._value < other

    def __int__(self):
        return int(self._value)

    def __float__(self):
        return float(self._value)

    def __bool__(self):
        return bool(self._value)

    __nonzero__ = __bool__


EXCLUDED_OPERATORS = set(
    util.CLASS_ATTRS
    + util.NEWCLASS_ATTRS
    + [
        "__name__",
        "__setattr__",
        "__getattr__",
        "__dict__",
        "__matmul__",
        "__imatmul__",
    ]
)


def dispatch_to_value(fn):
    def wrapper(self, other):
        return fn(self._value, other)
        pass

    return wrapper


""" This attaches all the literal operators to LiteralValue
 except for the reverse ones."""
for op in dir(operator):
    if op.startswith("__") and op not in EXCLUDED_OPERATORS:
        opfn = getattr(operator, op)
        setattr(LiteralValue, op, dispatch_to_value(opfn))


""" We also have to patch the reverse operators,
which aren't conveniently defined anywhere """
LiteralValue.__radd__ = lambda self, other: other + self._value
LiteralValue.__rsub__ = lambda self, other: other - self._value
LiteralValue.__rmul__ = lambda self, other: other * self._value
LiteralValue.__rtruediv__ = lambda self, other: other / self._value
LiteralValue.__rfloordiv__ = lambda self, other: other // self._value
LiteralValue.__rmod__ = lambda self, other: other % self._value
LiteralValue.__rdivmod__ = lambda self, other: divmod(other, self._value)
LiteralValue.__rpow__ = lambda self, other, modulo=None: pow(other, self._value, modulo)
LiteralValue.__rlshift__ = lambda self, other: other << self._value
LiteralValue.__rrshift__ = lambda self, other: other >> self._value
LiteralValue.__rand__ = lambda self, other: other & self._value
LiteralValue.__rxor__ = lambda self, other: other ^ self._value
LiteralValue.__ror__ = lambda self, other: other | self._value



import pytest
import json

import python_jsonschema_objects as pjs


def test_nested_arrays_work_fine():
    schema = {
        "$schema": "http://json-schema.org/draft-04/schema",
        "title": "Example1",
        "type": "object",
        "properties": {
            "name": {
                "type": "array",
                "items": {"type": "object", "properties": {"name": {"type": "string"}}},
            }
        },
    }

    ns1 = pjs.ObjectBuilder(schema).build_classes()
    j1 = ns1.Example1.from_json(
        json.dumps({"name": [{"value": "foo"}, {"value": "bar"}]})
    )
    assert j1.name[0].value == "foo"
    assert j1.name[1].value == "bar"


def test_top_level_arrays_are_converted_to_objects_properly():
    schema = {
        "$schema": "http://json-schema.org/draft-04/schema",
        "title": "Example2",
        "type": "array",
        "items": {"type": "object", "properties": {"name": {"type": "string"}}},
    }

    ns2 = pjs.ObjectBuilder(schema).build_classes()
    j2 = ns2.Example2.from_json(json.dumps([{"name": "foo"}, {"name": "bar"}]))
    assert not isinstance(j2[0], dict)  # Out[173]: {'name': 'foo'}
    assert j2[0].name == "foo"
    assert j2[1].name == "bar"



import json
import pytest
import python_jsonschema_objects as pjo

schema = """
{
  "$schema": "http://json-schema.org/draft-04/schema#",
  "title": "Default Test",
  "type": "object",
    "properties": {
        "p1": {
            "type": ["integer", "null"],
            "default": 0
        },
        "p2": {
            "type": ["integer", "null"],
            "default": null
        },
        "p3": {
            "type": ["integer", "null"]
        }
    }
}
"""


@pytest.fixture
def schema_json():
    return json.loads(schema)


@pytest.fixture
def ns(schema_json):
    builder = pjo.ObjectBuilder(schema_json)
    ns = builder.build_classes()
    return ns


def test_defaults_serialize_for_nullable_types(ns):
    thing1 = ns.DefaultTest()

    assert thing1.as_dict() == {"p1": 0, "p2": None}


def test_nullable_types_are_still_nullable(ns):
    thing1 = ns.DefaultTest()

    thing1.p1 = 10
    thing1.validate()
    assert thing1.as_dict() == {"p1": 10, "p2": None}

    thing1.p1 = None
    thing1.validate()
    assert thing1.as_dict() == {"p1": 0, "p2": None}


def test_null_types_without_defaults_do_not_serialize(ns):
    thing1 = ns.DefaultTest()

    thing1.p3 = 10
    thing1.validate()
    thing1.p1 = None
    thing1.validate()

    assert thing1.as_dict() == {"p1": 0, "p2": None, "p3": 10}



import pytest

import python_jsonschema_objects as pjo
import json


@pytest.fixture
def base_schema():
    return {
        "title": "example",
        "type": "object",
        "properties": {"foobar": {"type": "boolean"}},
        "patternProperties": {
            "^foo.*": {"type": "string"},
            "^bar.*": {"type": "integer"},
        },
    }


def test_standard_properties_take_precedence(base_schema):
    """foobar is a boolean, and it's a standard property,
    so we expect it will validate properly as a boolean,
    not using the patternProperty that matches it.
    """
    builder = pjo.ObjectBuilder(base_schema)
    ns = builder.build_classes()

    t = ns.Example(foobar=True)
    t.validate()

    with pytest.raises(pjo.ValidationError):
        # Try against the foo pattern
        x = ns.Example(foobar="hi")
        x.validate()


@pytest.mark.parametrize(
    "permit_addl,property,value,is_error",
    [
        (False, "foo", "hello", False),
        (False, "foobarro", "hello", False),
        (False, "foo", 24, True),
        (False, "barkeep", 24, False),
        (False, "barkeep", "John", True),
        (False, "extraprop", "John", True),
        (True, "extraprop", "John", False),
        # Test that the pattern props take precedence.
        # because these should validate against them, not the
        # additionalProperties that match
        (True, "foobar", True, False),
        (True, "foobar", "John", True),
        (True, "foobar", 24, True),
    ],
)
def test_pattern_properties_work(base_schema, permit_addl, property, value, is_error):

    base_schema["additionalProperties"] = permit_addl

    builder = pjo.ObjectBuilder(base_schema)
    ns = builder.build_classes()

    props = dict([(property, value)])

    if is_error:
        with pytest.raises(pjo.ValidationError):
            t = ns.Example(**props)
            t.validate()
    else:
        t = ns.Example(**props)
        t.validate()



import pytest

import python_jsonschema_objects as pjo


@pytest.fixture
def base_schema():
    return {
        "title": "example",
        "type": "object",
        "additionalProperties": False,
        "properties": {
            "dictLike": {"additionalProperties": {"type": "integer"}, "type": "object"}
        },
    }


def test_wrong_exception_protocolbase_getitem(base_schema):
    """
    to declare a dict like object in json-schema, we are supposed
    to declare it as an object of additional properties.
    When trying to use it as dict, for instance testing if a key is inside
    the dictionary, methods like __contains__ in the ProtocolBase expect
    __getitem__ to raise a KeyError. getitem calls __getattr__ without any
    exception handling, which raises an AttributeError (necessary for proper
    behaviour of getattr, for instance).
    Solution found is to handle AttributeError in getitem and to raise KeyError
    """
    builder = pjo.ObjectBuilder(base_schema)
    ns = builder.build_classes()

    t = ns.Example(dictLike={"a": 0, "b": 1})
    t.validate()
    assert "a" in t.dictLike
    assert "c" not in t.dictLike
    assert getattr(t, "not_present", None) is None

    with pytest.raises(AttributeError):
        assert "a" not in t.notAnAttribute


if __name__ == "__main__":
    test_wrong_exception_protocolbase_getitem(base_schema())



import pytest

from python_jsonschema_objects.validators import ValidationError
from python_jsonschema_objects.wrapper_types import ArrayWrapper


@pytest.mark.parametrize(
    "kwargs",
    [
        {},
        {"item_constraint": {"type": "string"}},
        {"item_constraint": [{"type": "string"}, {"type": "string"}]},
    ],
)
def test_ArrayValidator_initializer(kwargs):
    assert ArrayWrapper.create("hello", **kwargs)


def test_ArrayValidator_throws_error_if_not_classes_or_dicts():
    with pytest.raises(TypeError):
        ArrayWrapper.create("hello", item_constraint=["winner"])


def test_validate_basic_array_types():
    validator = ArrayWrapper.create("test", item_constraint={"type": "number"})

    instance = validator([1, 2, 3, 4])

    instance.validate()

    instance = validator([1, 2, "Hello"])
    with pytest.raises(ValidationError):
        instance.validate()


def test_validate_basic_tuple__types():
    validator = ArrayWrapper.create(
        "test", item_constraint=[{"type": "number"}, {"type": "number"}]
    )

    instance = validator([1, 2, 3, 4])

    instance.validate()

    instance = validator([1, "Hello"])
    with pytest.raises(ValidationError):
        instance.validate()


def test_validate_arrays_with_object_types(Person):
    validator = ArrayWrapper.create("test", item_constraint=Person)

    instance = validator([{"firstName": "winner", "lastName": "Smith"}])
    instance.validate()

    instance = validator(
        [{"firstName": "winner", "lastName": "Dinosaur"}, {"firstName": "BadMan"}]
    )
    with pytest.raises(ValidationError):
        instance.validate()


def test_validate_arrays_with_mixed_types(Person):

    validator = ArrayWrapper.create(
        "test", item_constraint=[Person, {"type": "number"}]
    )

    instance = validator([{"firstName": "winner", "lastName": "Dinosaur"}, "fried"])
    with pytest.raises(ValidationError):
        instance.validate()

    instance = validator([{"firstName": "winner", "lastName": "Dinosaur"}, 12324])
    instance.validate()


def test_validate_arrays_nested():

    validator = ArrayWrapper.create(
        "test", item_constraint={"type": "array", "items": {"type": "integer"}}
    )

    instance = validator([[1, 2, 4, 5], [1, 2, 4]])
    instance.validate()

    instance = validator([[1, 2, "h", 5], [1, 2, 4]])
    with pytest.raises(ValidationError):
        instance.validate()

    instance = validator([[1, 2, "h", 5], [1, 2, "4"]])
    with pytest.raises(ValidationError):
        instance.validate()


def test_validate_arrays_length():
    validator = ArrayWrapper.create("test", minItems=1, maxItems=3)

    instance = validator(range(1))
    instance.validate()

    instance = validator(range(2))
    instance.validate()

    instance = validator(range(3))
    instance.validate()

    instance = validator(range(4))
    with pytest.raises(ValidationError):
        instance.validate()

    instance = validator([])
    with pytest.raises(ValidationError):
        instance.validate()


def test_validate_arrays_uniqueness():
    validator = ArrayWrapper.create("test", uniqueItems=True)

    instance = validator([])
    instance.validate()

    instance = validator([1, 2, 3, 4])
    instance.validate()

    instance = validator([1, 2, 2, 4])
    with pytest.raises(ValidationError):
        instance.validate()



import json
import pytest
import python_jsonschema_objects as pjo

schema = {
    "$schema": "http://json-schema.org/draft-07/schema",
    "title": "myschema",
    "type": "object",
    "definitions": {
        "MainObject": {
            "title": "Main Object",
            "additionalProperties": False,
            "type": "object",
            "properties": {
                "location": {
                    "title": "location",
                    "type": ["object", "string"],
                    "oneOf": [
                        {"$ref": "#/definitions/UNIQUE_STRING"},
                        {"$ref": "#/definitions/Location"},
                    ],
                }
            },
        },
        "Location": {
            "title": "Location",
            "description": "A Location represents a span on a specific sequence.",
            "type": "object",
            "oneOf": [
                {"$ref": "#/definitions/Location1"},
                {"$ref": "#/definitions/Location2"},
            ],
            "discriminator": {"propertyName": "type"},
        },
        "Location1": {
            "additionalProperties": False,
            "type": "object",
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["Location1"],
                    "default": "Location1",
                }
            },
        },
        "Location2": {
            "additionalProperties": False,
            "type": "object",
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["Location2"],
                    "default": "Location2",
                }
            },
        },
        "UNIQUE_STRING": {
            "additionalProperties": False,
            "type": "string",
            "pattern": "^\\w[^:]*:.+$",
        },
    },
}


@pytest.fixture
def schema_json():
    return schema


def test_nested_oneofs_still_work(schema_json):
    builder = pjo.ObjectBuilder(schema_json)
    ns = builder.build_classes()

    obj1 = ns.MainObject(**{"location": {"type": "Location1"}})
    obj2 = ns.MainObject(**{"location": {"type": "Location2"}})
    obj3 = ns.MainObject(**{"location": "unique:12"})

    assert obj1.location.type == "Location1"
    assert obj2.location.type == "Location2"
    assert obj3.location == "unique:12"



import json
import pytest
import python_jsonschema_objects as pjo

schema = """
{
  "$schema": "http://json-schema.org/draft-04/schema#",
  "$id": "roundtrip.json",
  "type": "object",
  "properties":
  {
    "container": {"$ref": "#/definitions/config"}
  },
  "definitions": {
    "config": {
      "properties":{
        "something": {
          "allOf": [
            {
              "$ref": "#/definitions/params"
            },
            {
              "properties":{
                "parameters": {
                  "oneOf": [
                    {
                      "$ref": "#/definitions/parametersA"
                    },
                    {
                      "$ref": "#/definitions/parametersB"
                    }
                  ]
                }
              },
              "required": ["parameters"]
            }
          ]
        }
      }
    },
    "params": {
      "type": "object",
      "properties": {
        "param1": {
          "type": "string"
        },
        "param2": {
          "type": "string"
        }
      },
      "required": [
        "param1",
        "param2"
      ]
    },
    "parametersA": {
      "properties": {
        "name": {
          "type": "string"
        },
        "value": {
          "type": "string"
        }
      },
      "required": [
        "name",
        "value"
      ]
    },
    "parametersB": {
      "properties": {
        "something": {
          "type": "string"
        },
        "another": {
          "type": "string"
        }
      },
      "required": [
        "something",
        "another"
      ]
    }
  }
}
"""


@pytest.fixture
def schema_json():
    return json.loads(schema)


def test_roundtrip_oneof_serializer(schema_json):
    builder = pjo.ObjectBuilder(schema_json)
    namespace = builder.build_classes()

    data_config = """
    {
    "something": "a name",
    "another": "a value"
    }
    """
    paramsTypeB = namespace.Parametersb().from_json(data_config)
    somethingInstance = namespace.Something(param1="toto", param2="tata")
    somethingInstance.parameters = paramsTypeB

    json_object = somethingInstance.serialize()
    print(json_object)

    aNewsomething = namespace.Something.from_json(json_object)

    json_object2 = aNewsomething.serialize()
    assert json_object == json_object2



import pytest  # noqa
from jsonschema import RefResolver, Draft3Validator
from jsonschema._utils import load_schema, URIDict
import python_jsonschema_objects as pjo


def test_non_default_resolver_validator(markdown_examples):
    ms = URIDict()
    draft3 = load_schema("draft3")
    draft4 = load_schema("draft4")
    ms[draft3["id"]] = draft3
    ms[draft4["id"]] = draft4
    resolver_with_store = RefResolver(draft3["id"], draft3, ms)

    # 'Other' schema should be valid with draft3
    builder = pjo.ObjectBuilder(
        markdown_examples["Other"],
        resolver=resolver_with_store,
        validatorClass=Draft3Validator,
        resolved=markdown_examples,
    )
    klasses = builder.build_classes()
    a = klasses.Other(MyAddress="where I live")
    assert a.MyAddress == "where I live"



import six
import copy
import json


class lazy_format(object):
    __slots__ = ("fmt", "args", "kwargs")

    def __init__(self, fmt, *args, **kwargs):
        self.fmt = fmt
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        return self.fmt.format(*self.args, **self.kwargs)


def safe_issubclass(x, y):
    """Safe version of issubclass() that will not throw TypeErrors.

    Invoking issubclass('object', some-abc.meta instances) will result
    in the underlying implementation throwing TypeError's from trying to
    memoize the result- 'object' isn't a usable weakref target at that level.
    Unfortunately this gets exposed all the way up to our code; thus a
    'safe' version of the function."""
    try:
        return issubclass(x, y)
    except TypeError:
        return False


def coerce_for_expansion(mapping):
    """Given a value, make sure it is usable for f(**val) expansion.

    In py2.7, the value must be a dictionary- thus a as_dict() method
    will be invoked if available.  In py3k, the raw mapping is returned
    unmodified."""
    if six.PY2 and hasattr(mapping, "as_dict"):
        return mapping.as_dict()
    return mapping


class ProtocolJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        from python_jsonschema_objects import classbuilder
        from python_jsonschema_objects import wrapper_types

        if isinstance(obj, classbuilder.LiteralValue):
            return obj._value
        if isinstance(obj, wrapper_types.ArrayWrapper):
            return obj.for_json()
        if isinstance(obj, classbuilder.ProtocolBase):
            props = {}
            for raw, trans in six.iteritems(obj.__prop_names__):
                props[raw] = getattr(obj, trans)
                if props[raw] is None:
                    del props[raw]
            for raw, data in six.iteritems(obj._extended_properties):
                props[raw] = data
                if props[raw] is None:
                    del props[raw]
            return props
        else:
            return json.JSONEncoder.default(self, obj)


def propmerge(into, data_from):
    """ Merge JSON schema requirements into a dictionary """
    newprops = copy.deepcopy(into)

    for prop, propval in six.iteritems(data_from):
        if prop not in newprops:
            newprops[prop] = propval
            continue

        new_sp = newprops[prop]
        for subprop, spval in six.iteritems(propval):
            if subprop not in new_sp:
                new_sp[subprop] = spval

            elif subprop == "enum":
                new_sp[subprop] = set(spval) & set(new_sp[subprop])

            elif subprop == "type":
                if spval != new_sp[subprop]:
                    raise TypeError("Type cannot conflict in allOf'")

            elif subprop in ("minLength", "minimum"):
                new_sp[subprop] = new_sp[subprop] if new_sp[subprop] > spval else spval
            elif subprop in ("maxLength", "maximum"):
                new_sp[subprop] = new_sp[subprop] if new_sp[subprop] < spval else spval
            elif subprop == "multipleOf":
                if new_sp[subprop] % spval == 0:
                    new_sp[subprop] = spval
                else:
                    raise AttributeError("Cannot set conflicting multipleOf values")
            else:
                new_sp[subprop] = spval

        newprops[prop] = new_sp

    return newprops


def resolve_ref_uri(base, ref):
    if ref[0] == "#":
        # Local ref
        uri = base.rsplit("#", 1)[0] + ref
    else:
        uri = ref

    return uri


"""namespace module"""

__all__ = ("Namespace", "as_namespace")

from collections.abc import Mapping, Sequence


class _Dummy:
    pass


CLASS_ATTRS = dir(_Dummy)
NEWCLASS_ATTRS = dir(object)
del _Dummy


class Namespace(dict):

    """A dict subclass that exposes its items as attributes.

    Warning: Namespace instances do not have direct access to the
    dict methods.

    """

    def __init__(self, obj={}):
        dict.__init__(self, obj)

    def __dir__(self):
        return list(self)

    def __repr__(self):
        return "%s(%s)" % (type(self).__name__, super(dict, self).__repr__())

    def __getattribute__(self, name):
        try:
            return self[name]
        except KeyError:
            msg = "'%s' object has no attribute '%s'"
            raise AttributeError(msg % (type(self).__name__, name))

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        del self[name]

    # ------------------------
    # "copy constructors"

    @classmethod
    def from_object(cls, obj, names=None):
        if names is None:
            names = dir(obj)
        ns = {name: getattr(obj, name) for name in names}
        return cls(ns)

    @classmethod
    def from_mapping(cls, ns, names=None):
        if names:
            ns = {name: ns[name] for name in names}
        return cls(ns)

    @classmethod
    def from_sequence(cls, seq, names=None):
        if names:
            seq = {name: val for name, val in seq if name in names}
        return cls(seq)

    # ------------------------
    # static methods

    @staticmethod
    def hasattr(ns, name):
        try:
            object.__getattribute__(ns, name)
        except AttributeError:
            return False
        return True

    @staticmethod
    def getattr(ns, name):
        return object.__getattribute__(ns, name)

    @staticmethod
    def setattr(ns, name, value):
        return object.__setattr__(ns, name, value)

    @staticmethod
    def delattr(ns, name):
        return object.__delattr__(ns, name)


def as_namespace(obj, names=None):

    # functions
    if isinstance(obj, type(as_namespace)):
        obj = obj()

    # special cases
    if isinstance(obj, type):
        names = (name for name in dir(obj) if name not in CLASS_ATTRS)
        return Namespace.from_object(obj, names)
    if isinstance(obj, Mapping):
        return Namespace.from_mapping(obj, names)
    if isinstance(obj, Sequence):
        return Namespace.from_sequence(obj, names)

    # default
    return Namespace.from_object(obj, names)



import python_jsonschema_objects as pjs


def test_null_type():

    schema = {
        "$schema": "http://json-schema.org/draft-04/schema",
        "title": "Example1",
        "type": "object",
        "properties": {"foo": {"type": "null"}},
        "required": ["foo"],
    }

    ns1 = pjs.ObjectBuilder(schema).build_classes(strict=True)
    ns1.Example1(foo=None)


def test_null_type_one_of():

    schema = {
        "$schema": "http://json-schema.org/draft-04/schema",
        "title": "Example1",
        "type": "object",
        "properties": {"foo": {"oneOf": [{"type": "string"}, {"type": "null"}]}},
        "required": ["foo"],
    }

    ns1 = pjs.ObjectBuilder(schema).build_classes(strict=True)
    ns1.Example1(foo="bar")
    ns1.Example1(foo=None)



import python_jsonschema_objects.util as util
import python_jsonschema_objects.validators as validators
import python_jsonschema_objects.pattern_properties as pattern_properties
from python_jsonschema_objects.literals import LiteralValue

import copy
import collections
import itertools
import six
import sys

import logging

import python_jsonschema_objects.wrapper_types

logger = logging.getLogger(__name__)

logger.addHandler(logging.NullHandler())


# Long is no longer a thing in python3.x
if sys.version_info > (3,):
    long = int


class ProtocolBase(collections.abc.MutableMapping):
    """An instance of a class generated from the provided
    schema. All properties will be validated according to
    the definitions provided. However, whether or not all required
    properties have been provide will *not* be validated.

    Args:
        **props: Properties with which to populate the class object

    Returns:
        The class object populated with values

    Raises:
        validators.ValidationError: If any of the provided properties
            do not pass validation
    """

    __propinfo__ = {}
    __required__ = set()
    __has_default__ = set()
    __object_attr_list__ = set(["_properties", "_extended_properties"])

    def as_dict(self):
        """Return a dictionary containing the current values
        of the object.

        Returns:
            (dict): The object represented as a dictionary
        """
        out = {}
        for prop in self:
            propval = getattr(self, prop)

            if hasattr(propval, "for_json"):
                out[prop] = propval.for_json()
            elif isinstance(propval, list):
                out[prop] = [getattr(x, "for_json", lambda: x)() for x in propval]
            elif isinstance(propval, (ProtocolBase, LiteralValue)):
                out[prop] = propval.as_dict()
            elif propval is not None:
                out[prop] = propval

        return out

    def for_json(self):
        return self.as_dict()

    def __eq__(self, other):
        if not isinstance(other, ProtocolBase):
            return False

        return self.as_dict() == other.as_dict()

    def __str__(self):
        inverter = dict((v, k) for k, v in six.iteritems(self.__prop_names__))
        props = sorted(
            [
                "%s" % (inverter.get(k, k),)
                for k, v in itertools.chain(
                    six.iteritems(self._properties),
                    six.iteritems(self._extended_properties),
                )
            ]
        )
        return "<%s attributes: %s>" % (self.__class__.__name__, ", ".join(props))

    def __repr__(self):
        inverter = dict((v, k) for k, v in six.iteritems(self.__prop_names__))
        props = sorted(
            [
                "%s=%s" % (inverter.get(k, k), repr(v))
                for k, v in itertools.chain(
                    six.iteritems(self._properties),
                    six.iteritems(self._extended_properties),
                )
            ]
        )
        return "<%s %s>" % (self.__class__.__name__, " ".join(props))

    @classmethod
    def from_json(cls, jsonmsg):
        """Create an object directly from a JSON string.

        Applies general validation after creating the
        object to check whether all required fields are
        present.

        Args:
            jsonmsg (str): An object encoded as a JSON string

        Returns:
            An object of the generated type

        Raises:
            ValidationError: if `jsonmsg` does not match the schema
                `cls` was generated from
        """
        import json

        msg = json.loads(jsonmsg)
        obj = cls(**msg)
        obj.validate()
        return obj

    def __deepcopy__(self, memo):
        return self.__class__(**self.as_dict())

    def __new__(cls, **props):
        """Overridden to support oneOf, where we need to
        instantiate a different class depending on what
        value we've seen"""
        if getattr(cls, "__validation__", None) is None:
            new = super(ProtocolBase, cls).__new__
            if new is object.__new__:
                return new(cls)
            return new(cls, **props)

        valid_types = cls.__validation__.get("type", None)

        if valid_types is None or not isinstance(valid_types, list):
            new = super(ProtocolBase, cls).__new__
            if new is object.__new__:
                return new(cls)
            return new(cls, **props)

        obj = None
        validation_errors = []
        for klass in valid_types:
            logger.debug(
                util.lazy_format("Attempting to instantiate {0} as {1}", cls, klass)
            )
            try:
                obj = klass(**props)
                obj.validate()
            except validators.ValidationError as e:
                validation_errors.append((klass, e))
            else:
                break

        else:  # We got nothing
            raise validators.ValidationError(
                "Unable to instantiate any valid types: \n"
                "".join("{0}: {1}\n".format(k, e) for k, e in validation_errors)
            )

        return obj

    def __init__(self, **props):
        self._extended_properties = dict()
        self._properties = dict(
            zip(
                self.__prop_names__.values(),
                [None for x in six.moves.xrange(len(self.__prop_names__))],
            )
        )

        # To support defaults, we have to actually execute the constructors
        # but only for the ones that have defaults set.
        for name in self.__has_default__:
            if name not in props:
                default_value = copy.deepcopy(self.__propinfo__[name]["default"])
                logger.debug(
                    util.lazy_format("Initializing '{0}' to '{1}'", name, default_value)
                )
                setattr(self, name, default_value)

        for prop in props:
            try:
                logger.debug(
                    util.lazy_format(
                        "Setting value for '{0}' to {1}", prop, props[prop]
                    )
                )
                # Always set the property, even if None
                if props[prop] is not None:
                    setattr(self, prop, props[prop])
            except validators.ValidationError as e:
                import sys

                raise six.reraise(
                    type(e),
                    type(e)(
                        str(e)
                        + " \nwhile setting '{0}' in {1}".format(
                            prop, self.__class__.__name__
                        )
                    ),
                    sys.exc_info()[2],
                )

        if getattr(self, "__strict__", None):
            self.validate()

    def __setattr__(self, name, val):
        name = str(name)
        if name in self.__object_attr_list__:
            object.__setattr__(self, name, val)
        elif name in self.__propinfo__:
            # If its in __propinfo__, then it actually has a property defined.
            # The property does special validation, so we actually need to
            # run its setter. We get it from the class definition and call
            # it directly. XXX Heinous.
            prop = getattr(self.__class__, self.__prop_names__[name])
            prop.__set__(self, val)
        else:
            # This is an additional property of some kind
            try:
                val = self.__extensible__.instantiate(name, val)
            except Exception as e:
                raise validators.ValidationError(
                    "Attempted to set unknown property '{0}': {1} ".format(name, e)
                )

            self._extended_properties[name] = val

    """ Implement collections.MutableMapping methods """

    def __iter__(self):
        import itertools

        return itertools.chain(
            six.iterkeys(self._extended_properties), six.iterkeys(self._properties)
        )

    def __len__(self):
        return len(self._extended_properties) + len(self._properties)

    def __getitem__(self, key):
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(key)

    def __getattr__(self, name):
        name = str(name)
        if name in self.__prop_names__:
            raise AttributeError(name)
        if name in self._extended_properties:
            return self._extended_properties[name]
        raise AttributeError(
            "{0} is not a valid property of {1}".format(name, self.__class__.__name__)
        )

    def __setitem__(self, key, val):
        key = str(key)
        return setattr(self, key, val)

    def __delitem__(self, key):
        key = str(key)
        return delattr(self, key)

    def __delattr__(self, name):
        name = str(name)
        if name in self._extended_properties:
            del self._extended_properties[name]
            return

        if name in self.__prop_names__:
            prop = getattr(self.__class__, self.__prop_names__[name])
            prop.__delete__(self)
            return

        return object.__delattr__(self, name)

    @classmethod
    def propinfo(cls, propname):
        if propname not in cls.__propinfo__:
            return {}
        return cls.__propinfo__[propname]

    def serialize(self, **opts):
        self.validate()
        enc = util.ProtocolJSONEncoder(**opts)
        return enc.encode(self)

    def validate(self):
        """Applies all defined validation to the current
        state of the object, and raises an error if
        they are not all met.

        Raises:
            ValidationError: if validations do not pass
        """

        missing = self.missing_property_names()

        if len(missing) > 0:
            raise validators.ValidationError(
                "'{0}' are required attributes for {1}".format(
                    missing, self.__class__.__name__
                )
            )

        for prop, val in six.iteritems(self._properties):
            if val is None:
                continue

            if isinstance(val, ProtocolBase):
                val.validate()
            elif getattr(val, "isLiteralClass", None) is True:
                val.validate()
            elif isinstance(val, list):
                for subval in val:
                    subval.validate()
            else:
                # This object is of the wrong type, but just try setting it
                # The property setter will enforce its correctness
                # and handily coerce its type at the same time
                setattr(self, prop, val)

        return True

    def missing_property_names(self):
        """
        Returns a list of properties which are required and missing.

        Properties are excluded from this list if they are allowed to be null.

        :return: list of missing properties.
        """

        propname = lambda x: self.__prop_names__[x]
        missing = []
        for x in self.__required__:

            # Allow the null type
            propinfo = self.propinfo(propname(x))
            null_type = False
            if "type" in propinfo:
                type_info = propinfo["type"]
                null_type = (
                    type_info == "null"
                    or isinstance(type_info, (list, tuple))
                    and "null" in type_info
                )
            elif "oneOf" in propinfo:
                for o in propinfo["oneOf"]:
                    type_info = o.get("type")
                    if (
                        type_info
                        and type_info == "null"
                        or isinstance(type_info, (list, tuple))
                        and "null" in type_info
                    ):
                        null_type = True
                        break

            if (propname(x) not in self._properties and null_type) or (
                self._properties[propname(x)] is None and not null_type
            ):
                missing.append(x)

        return missing


class TypeRef(object):
    def __init__(self, ref_uri, resolved):
        self._resolved = resolved
        self._ref_uri = ref_uri
        self._class = None
        self.__doc__ = "Reference to {}".format(ref_uri)

    @property
    def ref_class(self):
        if self._class is None:
            self._class = self._resolved.get(self._ref_uri)
        return self._class

    def __call__(self, *args, **kwargs):
        cls = self.ref_class
        return cls(*args, **kwargs)

    def __str__(self):
        return self.__doc__

    def __repr__(self):
        return "<{}>".format(self.__doc__)


class TypeProxy(object):
    slots = ("__title__", "_types")

    def __init__(self, types, title=None):
        self.__title__ = title
        self._types = types

    def from_json(self, jsonmsg):
        import json

        msg = json.loads(jsonmsg)
        obj = self(**msg)
        obj.validate()
        return obj

    def __call__(self, *a, **kw):
        validation_errors = []
        valid_types = self._types
        for klass in valid_types:
            logger.debug(
                util.lazy_format(
                    "Attempting to instantiate {0} as {1}", self.__class__, klass
                )
            )
            try:
                obj = klass(*a, **kw)
                obj.validate()
            except TypeError as e:
                validation_errors.append((klass, e))
            except validators.ValidationError as e:
                validation_errors.append((klass, e))
            else:
                return obj

        else:  # We got nothing
            raise validators.ValidationError(
                "Unable to instantiate any valid types: \n"
                "".join("{0}: {1}\n".format(k, e) for k, e in validation_errors)
            )


class ClassBuilder(object):
    def __init__(self, resolver):
        self.resolver = resolver
        self.resolved = {}
        self.under_construction = set()

    def expand_references(self, source_uri, iterable):
        """Give an iterable of jsonschema descriptors, expands any
        of them that are $ref objects, and otherwise leaves them alone.
        """
        pp = []
        for elem in iterable:
            if "$ref" in elem:
                pp.append(self.resolve_type(elem["$ref"], source_uri))
            else:
                pp.append(elem)

        return pp

    def resolve_type(self, ref, source):
        """ Return a resolved type for a URI, potentially constructing one if necessary"""
        uri = util.resolve_ref_uri(self.resolver.resolution_scope, ref)
        if uri in self.resolved:
            return self.resolved[uri]

        elif uri in self.under_construction:
            logger.debug(
                util.lazy_format(
                    "Using a TypeRef to avoid a cyclic reference for {0} -> {1} ",
                    uri,
                    source,
                )
            )
            return TypeRef(uri, self.resolved)
        else:
            logger.debug(
                util.lazy_format(
                    "Resolving direct reference object {0} -> {1}", source, uri
                )
            )
            with self.resolver.resolving(ref) as resolved:
                self.resolved[uri] = self.construct(uri, resolved, (ProtocolBase,))
                return self.resolved[uri]

    def construct(self, uri, *args, **kw):
        """ Wrapper to debug things """
        logger.debug(util.lazy_format("Constructing {0}", uri))
        if ("override" not in kw or kw["override"] is False) and uri in self.resolved:
            logger.debug(util.lazy_format("Using existing {0}", uri))
            assert self.resolved[uri] is not None
            return self.resolved[uri]
        else:
            ret = self._construct(uri, *args, **kw)
        logger.debug(util.lazy_format("Constructed {0}", ret))

        return ret

    def _construct(self, uri, clsdata, parent=(ProtocolBase,), **kw):

        if "anyOf" in clsdata:
            raise NotImplementedError("anyOf is not supported as bare property")

        elif "oneOf" in clsdata:
            """If this object itself has a 'oneOf' designation,
            then construct a TypeProxy.
            """
            klasses = self.construct_objects(clsdata["oneOf"], uri)

            logger.debug(
                util.lazy_format("Designating {0} as TypeProxy for {1}", uri, klasses)
            )
            self.resolved[uri] = TypeProxy(klasses, title=clsdata.get("title"))
            return self.resolved[uri]

        elif "allOf" in clsdata:
            potential_parents = self.expand_references(uri, clsdata["allOf"])
            parents = []
            for p in potential_parents:
                if isinstance(p, dict):
                    # This is additional constraints
                    clsdata.update(p)
                elif util.safe_issubclass(p, ProtocolBase):
                    parents.append(p)

            self.resolved[uri] = self._build_object(uri, clsdata, parents, **kw)
            return self.resolved[uri]

        elif "$ref" in clsdata:

            if "type" in clsdata and util.safe_issubclass(
                clsdata["type"], (ProtocolBase, LiteralValue)
            ):
                # It's possible that this reference was already resolved, in which
                # case it will have its type parameter set
                logger.debug(
                    util.lazy_format(
                        "Using previously resolved type "
                        "(with different URI) for {0}",
                        uri,
                    )
                )
                self.resolved[uri] = clsdata["type"]
            elif uri in self.resolved:
                logger.debug(
                    util.lazy_format("Using previously resolved object for {0}", uri)
                )
            else:
                ref = clsdata["$ref"]
                typ = self.resolve_type(ref, uri)
                self.resolved[uri] = typ

            return self.resolved[uri]

        elif clsdata.get("type") == "array" and "items" in clsdata:
            clsdata_copy = {}
            clsdata_copy.update(clsdata)
            self.resolved[
                uri
            ] = python_jsonschema_objects.wrapper_types.ArrayWrapper.create(
                uri,
                item_constraint=clsdata_copy.pop("items"),
                classbuilder=self,
                **clsdata_copy
            )
            return self.resolved[uri]

        elif isinstance(clsdata.get("type"), list):
            types = []
            for i, item_detail in enumerate(clsdata["type"]):
                subdata = {k: v for k, v in six.iteritems(clsdata) if k != "type"}
                subdata["type"] = item_detail
                types.append(self._build_literal(uri + "_%s" % i, subdata))

            self.resolved[uri] = TypeProxy(types)
            return self.resolved[uri]

        elif (
            clsdata.get("type", None) == "object"
            or clsdata.get("properties", None) is not None
            or clsdata.get("additionalProperties", False)
        ):
            self.resolved[uri] = self._build_object(uri, clsdata, parent, **kw)
            return self.resolved[uri]
        elif clsdata.get("type") in ("integer", "number", "string", "boolean", "null"):
            self.resolved[uri] = self._build_literal(uri, clsdata)
            return self.resolved[uri]
        elif "enum" in clsdata:
            obj = self._build_literal(uri, clsdata)
            self.resolved[uri] = obj
            return obj

        elif "type" in clsdata and util.safe_issubclass(clsdata["type"], ProtocolBase):
            self.resolved[uri] = clsdata.get("type")
            return self.resolved[uri]
        else:
            raise NotImplementedError(
                "Unable to parse schema object '{0}' with "
                "no type and no reference".format(clsdata)
            )

    def _build_literal(self, nm, clsdata):
        """@todo: Docstring for _build_literal

        :nm: @todo
        :clsdata: @todo
        :returns: @todo

        """
        cls = type(
            str(nm),
            tuple((LiteralValue,)),
            {
                "__propinfo__": {
                    "__literal__": clsdata,
                    "__title__": clsdata.get("title"),
                    "__default__": clsdata.get("default"),
                }
            },
        )

        return cls

    def _build_object(self, nm, clsdata, parents, **kw):
        logger.debug(util.lazy_format("Building object {0}", nm))

        # To support circular references, we tag objects that we're
        # currently building as "under construction"
        self.under_construction.add(nm)

        props = {}
        defaults = set()

        properties = {}
        for p in parents:
            properties = util.propmerge(properties, p.__propinfo__)

        if "properties" in clsdata:
            properties = util.propmerge(properties, clsdata["properties"])

        name_translation = {}

        for prop, detail in properties.items():
            logger.debug(util.lazy_format("Handling property {0}.{1}", nm, prop))
            properties[prop]["raw_name"] = prop
            name_translation[prop] = prop.replace("@", "")
            prop = name_translation[prop]

            # Set default value, even if None
            if "default" in detail:
                defaults.add(prop)

            if detail.get("type", None) == "object":
                uri = "{0}/{1}_{2}".format(nm, prop, "<anonymous>")
                self.resolved[uri] = self.construct(uri, detail, (ProtocolBase,))

                props[prop] = make_property(
                    prop, {"type": self.resolved[uri]}, self.resolved[uri].__doc__
                )
                properties[prop]["type"] = self.resolved[uri]

            elif "type" not in detail and "$ref" in detail:
                ref = detail["$ref"]
                typ = self.resolve_type(ref, ".".join([nm, prop]))

                props[prop] = make_property(prop, {"type": typ}, typ.__doc__)
                properties[prop]["$ref"] = ref
                properties[prop]["type"] = typ

            elif "oneOf" in detail:
                potential = self.expand_references(nm, detail["oneOf"])
                logger.debug(
                    util.lazy_format("Designating {0} as oneOf {1}", prop, potential)
                )
                desc = detail["description"] if "description" in detail else ""
                props[prop] = make_property(prop, {"type": potential}, desc)

            elif "type" in detail and detail["type"] == "array":
                if "items" in detail and isinstance(detail["items"], dict):
                    if "$ref" in detail["items"]:
                        typ = self.resolve_type(detail["items"]["$ref"], nm)
                        constraints = copy.copy(detail)
                        constraints["strict"] = kw.get("strict")
                        propdata = {
                            "type": "array",
                            "validator": python_jsonschema_objects.wrapper_types.ArrayWrapper.create(
                                nm, item_constraint=typ, **constraints
                            ),
                        }

                    else:
                        uri = "{0}/{1}_{2}".format(nm, prop, "<anonymous_field>")
                        try:
                            if "oneOf" in detail["items"]:
                                typ = TypeProxy(
                                    self.construct_objects(
                                        detail["items"]["oneOf"], uri
                                    )
                                )
                            else:
                                typ = self.construct(uri, detail["items"])

                            constraints = copy.copy(detail)
                            constraints["strict"] = kw.get("strict")
                            propdata = {
                                "type": "array",
                                "validator": python_jsonschema_objects.wrapper_types.ArrayWrapper.create(
                                    uri, item_constraint=typ, **constraints
                                ),
                            }

                        except NotImplementedError:
                            typ = detail["items"]
                            constraints = copy.copy(detail)
                            constraints["strict"] = kw.get("strict")
                            propdata = {
                                "type": "array",
                                "validator": python_jsonschema_objects.wrapper_types.ArrayWrapper.create(
                                    uri, item_constraint=typ, **constraints
                                ),
                            }

                    props[prop] = make_property(prop, propdata, typ.__doc__)
                elif "items" in detail:
                    typs = []
                    for i, elem in enumerate(detail["items"]):
                        uri = "{0}/{1}/<anonymous_{2}>".format(nm, prop, i)
                        typ = self.construct(uri, elem)
                        typs.append(typ)

                    props[prop] = make_property(prop, {"type": typs})

            else:
                desc = detail["description"] if "description" in detail else ""
                uri = "{0}/{1}".format(nm, prop)
                typ = self.construct(uri, detail)

                props[prop] = make_property(prop, {"type": typ}, desc)

        props["__extensible__"] = pattern_properties.ExtensibleValidator(
            nm, clsdata, self
        )

        props["__prop_names__"] = name_translation

        props["__propinfo__"] = properties
        required = set.union(*[p.__required__ for p in parents])

        if "required" in clsdata:
            for prop in clsdata["required"]:
                required.add(prop)

        invalid_requires = [req for req in required if req not in props["__propinfo__"]]
        if len(invalid_requires) > 0:
            raise validators.ValidationError(
                "Schema Definition Error: {0} schema requires "
                "'{1}', but properties are not defined".format(nm, invalid_requires)
            )

        props["__required__"] = required
        props["__has_default__"] = defaults
        if required and kw.get("strict"):
            props["__strict__"] = True

        props["__title__"] = clsdata.get("title")
        cls = type(str(nm.split("/")[-1]), tuple(parents), props)
        self.under_construction.remove(nm)

        return cls

    def construct_objects(self, oneOfList, uri):
        return [
            self.construct(uri + "_%s" % i, item_detail)
            if "$ref" not in item_detail
            else self.resolve_type(item_detail["$ref"], uri + "_%s" % i)
            for i, item_detail in enumerate(oneOfList)
        ]


def make_property(prop, info, desc=""):
    from . import descriptors

    prop = descriptors.AttributeDescriptor(prop, info, desc)
    return prop



import python_jsonschema_objects as pjs
import pytest


@pytest.fixture
def testclass():
    builder = pjs.ObjectBuilder({"title": "Test", "type": "object"})
    ns = builder.build_classes()
    return ns.Test()


def test_extra_properties_can_be_deleted_with_item_syntax(testclass):
    # Deletion before setting should throw attribute errors
    with pytest.raises(AttributeError):
        del testclass["foo"]

    testclass.foo = 42
    assert testclass.foo == 42
    del testclass["foo"]

    # Etestclasstra properties not set should raise AttributeErrors when accessed
    with pytest.raises(AttributeError):
        testclass.foo


def test_extra_properties_can_be_deleted_with_attribute_syntax(testclass):
    # Deletion before setting should throw attribute errors
    with pytest.raises(AttributeError):
        del testclass.foo

    testclass.foo = 42
    assert testclass.foo == 42
    del testclass.foo

    # Extra properties not set should raise AttributeErrors when accessed
    with pytest.raises(AttributeError):
        testclass.foo


def test_extra_properties_can_be_deleted_directly(testclass):
    # Deletion before setting should throw attribute errors
    with pytest.raises(AttributeError):
        delattr(testclass, "foo")

    testclass.foo = 42
    assert testclass.foo == 42
    delattr(testclass, "foo")

    # Extra properties not set should raise AttributeErrors when accessed
    with pytest.raises(AttributeError):
        testclass.foo


def test_unrequired_real_properties_arent_really_deleted(Person):
    p = Person(age=20)

    assert p.age == 20
    del p.age
    assert p.age == None

    p.age = 20
    assert p.age == 20
    delattr(p, "age")
    assert p.age == None

    p.age = 20
    assert p.age == 20
    del p["age"]
    assert p.age == None


def test_required_real_properties_throw_attributeerror_on_delete(Person):
    p = Person(firstName="Fred")

    assert p.firstName == "Fred"
    with pytest.raises(AttributeError):
        del p.firstName

    p.firstName = "Fred"
    assert p.firstName == "Fred"
    with pytest.raises(AttributeError):
        delattr(p, "firstName")

    p.firstName = "Fred"
    assert p.firstName == "Fred"
    with pytest.raises(AttributeError):
        del p["firstName"]



from . import validators, util, wrapper_types
from .classbuilder import ProtocolBase, TypeProxy, TypeRef


class AttributeDescriptor(object):
    """ Provides property access for constructed class properties """

    def __init__(self, prop, info, desc=""):
        self.prop = prop
        self.info = info
        self.desc = desc

    def __doc__(self):
        return self.desc

    def __get__(self, obj, owner=None):
        if obj is None and owner is not None:
            return self

        try:
            return obj._properties[self.prop]
        except KeyError:
            raise AttributeError("No such attribute")

    def __set__(self, obj, val):
        info = self.info
        if isinstance(info["type"], (list, tuple)):
            ok = False
            errors = []
            type_checks = []

            for typ in info["type"]:
                if not isinstance(typ, dict):
                    type_checks.append(typ)
                    continue
                typ = next(
                    t for n, t in validators.SCHEMA_TYPE_MAPPING if typ["type"] == n
                )
                if typ is None:
                    typ = type(None)
                if isinstance(typ, (list, tuple)):
                    type_checks.extend(typ)
                else:
                    type_checks.append(typ)

            for typ in type_checks:
                if not isinstance(typ, TypeProxy) and isinstance(val, typ):
                    ok = True
                    break
                elif hasattr(typ, "isLiteralClass"):
                    try:
                        validator = typ(val)
                        validator.validate()
                    except Exception as e:
                        errors.append("Failed to coerce to '{0}': {1}".format(typ, e))
                        pass
                    else:
                        ok = True
                        break
                elif util.safe_issubclass(typ, ProtocolBase):
                    # force conversion- thus the val rather than validator assignment
                    try:
                        val = typ(**util.coerce_for_expansion(val))
                        val.validate()
                    except Exception as e:
                        errors.append("Failed to coerce to '{0}': {1}".format(typ, e))
                        pass
                    else:
                        ok = True
                        break
                elif util.safe_issubclass(typ, wrapper_types.ArrayWrapper):
                    try:
                        val = typ(val)
                        val.validate()
                    except Exception as e:
                        errors.append("Failed to coerce to '{0}': {1}".format(typ, e))
                        pass
                    else:
                        ok = True
                        break
                elif isinstance(typ, TypeProxy):
                    try:
                        val = typ(**util.coerce_for_expansion(val))
                        val.validate()
                    except Exception as e:
                        errors.append("Failed to coerce to '{0}': {1}".format(typ, e))
                        pass
                    else:
                        ok = True
                        break

            if not ok:
                errstr = "\n".join(errors)
                raise validators.ValidationError(
                    "Object must be one of {0}: \n{1}".format(info["type"], errstr)
                )

        elif info["type"] == "array":
            val = info["validator"](val)
            val.validate()

        elif util.safe_issubclass(info["type"], wrapper_types.ArrayWrapper):
            # An array type may have already been converted into an ArrayValidator
            val = info["type"](val)
            val.validate()

        elif getattr(info["type"], "isLiteralClass", False) is True:
            if not isinstance(val, info["type"]):
                validator = info["type"](val)
                validator.validate()
                if validator._value is not None:
                    # This allows setting of default Literal values
                    val = validator

        elif util.safe_issubclass(info["type"], ProtocolBase):
            if not isinstance(val, info["type"]):
                val = info["type"](**util.coerce_for_expansion(val))

            val.validate()

        elif isinstance(info["type"], TypeProxy):
            val = info["type"](val)

        elif isinstance(info["type"], TypeRef):
            if not isinstance(val, info["type"].ref_class):
                val = info["type"](**val)

            val.validate()

        elif info["type"] is None:
            # This is the null value
            if val is not None:
                raise validators.ValidationError("None is only valid value for null")

        else:
            raise TypeError("Unknown object type: '{0}'".format(info["type"]))

        obj._properties[self.prop] = val

    def __delete__(self, obj):
        prop = self.prop
        if prop in obj.__required__:
            raise AttributeError("'%s' is required" % prop)
        else:
            obj._properties[prop] = None



import python_jsonschema_objects as pjs


def test_nested_arrays_work_fine():
    schema = {
        "title": "Example Schema",
        "type": "object",
        "properties": {
            "a": {"type": "array", "items": {"type": "string"}, "default": []}
        },
    }

    ns1 = pjs.ObjectBuilder(schema).build_classes()

    foo = ns1.ExampleSchema()
    foo.a.append("foo")
    print(foo.for_json())

    assert foo.a == ["foo"]

    bar = ns1.ExampleSchema()
    bar.a.append("bar")
    print(bar.for_json())

    assert bar.a == ["bar"]



import logging

import six

logger = logging.getLogger(__name__)

SCHEMA_TYPE_MAPPING = (
    ("array", list),
    ("boolean", bool),
    ("integer", six.integer_types),
    ("number", six.integer_types + (float,)),
    ("null", type(None)),
    ("string", six.string_types),
    ("object", dict),
)
"""Sequence of schema type mappings to be checked in precedence order."""


class ValidationError(Exception):
    pass


class ValidatorRegistry(object):
    def __init__(self):
        self.registry = {}

    def register(self, name=None):
        def f(functor):
            self.registry[name if name is not None else functor.__name__] = functor
            return functor

        return f

    def __call__(self, name):
        return self.registry.get(name)


registry = ValidatorRegistry()


@registry.register()
def multipleOf(param, value, _):
    quot, rem = divmod(value, param)
    if rem != 0:
        raise ValidationError("{0} is not a multiple of {1}".format(value, param))


@registry.register()
def enum(param, value, _):
    if value not in param:
        raise ValidationError("{0} is not one of {1}".format(value, param))


@registry.register()
def minimum(param, value, type_data):
    exclusive = type_data.get("exclusiveMinimum")
    if exclusive:
        if value <= param:
            raise ValidationError(
                "{0} is less than or equal to {1}".format(value, param)
            )
    elif value < param:
        raise ValidationError("{0} is less than {1}".format(value, param))


@registry.register()
def maximum(param, value, type_data):
    exclusive = type_data.get("exclusiveMaximum")
    if exclusive:
        if value >= param:
            raise ValidationError(
                "{0} is greater than or equal to {1}".format(value, param)
            )
    elif value > param:
        raise ValidationError("{0} is greater than {1}".format(value, param))


@registry.register()
def maxLength(param, value, _):
    if len(value) > param:
        raise ValidationError("{0} is longer than {1} characters".format(value, param))


@registry.register()
def minLength(param, value, _):
    if len(value) < param:
        raise ValidationError("{0} is fewer than {1} characters".format(value, param))


@registry.register()
def pattern(param, value, _):
    import re

    match = re.search(param, value)
    if not match:
        raise ValidationError("{0} does not match {1}".format(value, param))


try:
    from jsonschema import FormatChecker
except ImportError:
    pass
else:

    @registry.register()
    def format(param, value, _):
        if not FormatChecker().conforms(value, param):
            raise ValidationError(
                "'{0}' is not formatted as a {1}".format(value, param)
            )


type_registry = ValidatorRegistry()


@type_registry.register(name="boolean")
def check_boolean_type(param, value, _):
    if not isinstance(value, bool):
        raise ValidationError("{0} is not a boolean".format(value))


@type_registry.register(name="integer")
def check_integer_type(param, value, _):
    if not isinstance(value, int) or isinstance(value, bool):
        raise ValidationError("{0} is not an integer".format(value))


@type_registry.register(name="number")
def check_number_type(param, value, _):
    if not isinstance(value, six.integer_types + (float,)) or isinstance(value, bool):
        raise ValidationError("{0} is neither an integer nor a float".format(value))


@type_registry.register(name="null")
def check_null_type(param, value, _):
    if value is not None:
        raise ValidationError("{0} is not None".format(value))


@type_registry.register(name="string")
def check_string_type(param, value, _):
    if not isinstance(value, six.string_types):
        raise ValidationError("{0} is not a string".format(value))


@type_registry.register(name="array")
def check_array_type(param, value, _):
    if not isinstance(value, list):
        raise ValidationError("{0} is not an array".format(value))


@type_registry.register(name="object")
def check_object_type(param, value, _):
    from python_jsonschema_objects.classbuilder import ProtocolBase

    if not isinstance(value, (dict, ProtocolBase)):
        raise ValidationError(
            "{0} is not an object (neither dict nor ProtocolBase)".format(value)
        )


@registry.register(name="type")
def check_type(param, value, type_data):
    type_check = type_registry(param)
    if type_check is None:
        raise ValidationError("{0} is an invalid type".format(value))
    type_check(param, value, type_data)



import pytest  # noqa
import python_jsonschema_objects as pjo


def test_regression_156(markdown_examples):
    builder = pjo.ObjectBuilder(
        markdown_examples["MultipleObjects"], resolved=markdown_examples
    )
    classes = builder.build_classes(named_only=True)

    er = classes.ErrorResponse(message="Danger!", status=99)
    vgr = classes.VersionGetResponse(local=False, version="1.2.3")

    # round-trip serialize-deserialize into named classes
    classes.ErrorResponse.from_json(er.serialize())
    classes.VersionGetResponse.from_json(vgr.serialize())

    # round-trip serialize-deserialize into class defined with `oneOf`
    classes.Multipleobjects.from_json(er.serialize())
    classes.Multipleobjects.from_json(vgr.serialize())


def test_toplevel_oneof_gets_a_name(markdown_examples):
    builder = pjo.ObjectBuilder(
        markdown_examples["MultipleObjects"], resolved=markdown_examples
    )
    classes = builder.build_classes(named_only=True)
    assert classes.Multipleobjects.__title__ is not None



import pytest

import python_jsonschema_objects as pjo


@pytest.fixture
def test_instance():
    schema = {
        "title": "Example",
        "properties": {
            "stringProp": {"type": "string"},
            "arrayProp": {"type": "array", "items": {"type": "string"}},
        },
    }

    builder = pjo.ObjectBuilder(schema)
    ns = builder.build_classes()
    instance = ns.Example(
        stringProp="This seems fine", arrayProp=["these", "are", "problematic"]
    )
    return instance


def test_string_properties_compare_to_strings(test_instance):
    test = test_instance.stringProp == "This seems fine"
    assert test


def test_arrays_of_strings_compare_to_strings(test_instance):
    test = test_instance.arrayProp == ["these", "are", "problematic"]
    assert test


def test_arrays_can_be_extended(test_instance):
    test_instance.arrayProp.extend(["...", "maybe", "not"])
    test = test_instance.arrayProp == [
        "these",
        "are",
        "problematic",
        "...",
        "maybe",
        "not",
    ]
    assert test


def test_array_elements_compare_to_types(test_instance):
    elem = test_instance.arrayProp[0]
    test = elem == "these"
    assert test


def test_repr_shows_property_values(test_instance):
    expected = "<Literal<str> these>"
    assert repr(test_instance.arrayProp[0]) == expected


def test_str_shows_just_strings(test_instance):
    test = str(test_instance.arrayProp[0])
    assert test == "these"



import six
import re
import python_jsonschema_objects.validators as validators
import python_jsonschema_objects.util as util
from python_jsonschema_objects.literals import MakeLiteral

import collections

import logging

import python_jsonschema_objects.wrapper_types

logger = logging.getLogger(__name__)

PatternDef = collections.namedtuple("PatternDef", "pattern schema_type")


class ExtensibleValidator(object):
    def __init__(self, name, schemadef, builder):
        import python_jsonschema_objects.classbuilder as cb

        self._pattern_types = []
        self._additional_type = True

        addlProp = schemadef.get("additionalProperties", True)

        if addlProp is False:
            self._additional_type = False
        elif addlProp is True:
            self._additional_type = True
        else:
            if "$ref" in addlProp:
                typ = builder.resolve_type(addlProp["$ref"], name)
            else:
                uri = "{0}/{1}_{2}".format(
                    name, "<additionalProperties>", "<anonymous>"
                )
                builder.resolved[uri] = builder.construct(
                    uri, addlProp, (cb.ProtocolBase,)
                )
                typ = builder.resolved[uri]

            self._additional_type = typ

        for pattern, typedef in six.iteritems(schemadef.get("patternProperties", {})):
            if "$ref" in typedef:
                typ = builder.resolve_type(typedef["$ref"], name)
            else:
                uri = "{0}/{1}_{2}".format(name, "<patternProperties>", pattern)

                builder.resolved[uri] = builder.construct(
                    uri, typedef, (cb.ProtocolBase,)
                )
                typ = builder.resolved[uri]

            self._pattern_types.append(
                PatternDef(pattern=re.compile(pattern), schema_type=typ)
            )

    def _make_type(self, typ, val):
        import python_jsonschema_objects.classbuilder as cb

        if getattr(typ, "isLiteralClass", None) is True:
            return typ(val)

        if util.safe_issubclass(typ, cb.ProtocolBase):
            return typ(**util.coerce_for_expansion(val))

        if util.safe_issubclass(
            typ, python_jsonschema_objects.wrapper_types.ArrayWrapper
        ):
            return typ(val)

        raise validators.ValidationError(
            "additionalProperty type {0} was neither a literal "
            "nor a schema wrapper: {1}".format(typ, val)
        )

    def instantiate(self, name, val):
        import python_jsonschema_objects.classbuilder as cb

        for p in self._pattern_types:
            if p.pattern.search(name):
                logger.debug(
                    "Found patternProperties match: %s %s" % (p.pattern.pattern, name)
                )
                return self._make_type(p.schema_type, val)

        if self._additional_type is True:

            valtype = [
                k
                for k, t in validators.SCHEMA_TYPE_MAPPING
                if t is not None and isinstance(val, t)
            ]
            valtype = valtype[0]
            return MakeLiteral(name, valtype, val)

        elif isinstance(self._additional_type, type):
            return self._make_type(self._additional_type, val)

        raise validators.ValidationError(
            "additionalProperties not permitted " "and no patternProperties specified"
        )



import pytest

import python_jsonschema_objects as pjo


@pytest.fixture
def test_class():
    schema = {
        "title": "Example",
        "properties": {
            "claimed_by": {
                "$id": "claimed",
                "type": ["string", "integer", "null"],
                "description": "Robots Only. The human agent that has claimed this robot.",
            }
        },
    }

    builder = pjo.ObjectBuilder(schema)
    ns = builder.build_classes()
    return ns


@pytest.mark.parametrize("value", ["Hi", 4, None])
def test_properties_can_have_multiple_types(test_class, value):
    test_class.Example(claimed_by=value)


@pytest.mark.parametrize("value", [2.4])
def test_multiply_typed_properties_still_validate(test_class, value):
    with pytest.raises(pjo.ValidationError):
        test_class.Example(claimed_by=value)



import python_jsonschema_objects as pjs
import python_jsonschema_objects.wrapper_types


def test_limited_validation(mocker):
    schema = {
        "title": "Example Schema",
        "type": "object",
        "properties": {
            "a": {"type": "array", "items": {"type": "string"}, "default": []}
        },
    }

    ob = pjs.ObjectBuilder(schema)
    ns1 = ob.build_classes()
    validator = ns1.ExampleSchema.a.info["validator"]([])
    validate_items = mocker.patch.object(
        validator, "validate_items", side_effect=validator.validate_items
    )
    validate = mocker.patch.object(
        validator, "validate", side_effect=validator.validate
    )
    mocker.patch.dict(
        ns1.ExampleSchema.a.info,
        {"validator": mocker.MagicMock(return_value=validator)},
    )

    foo = ns1.ExampleSchema()
    # We expect validation to be called on creation
    assert validate_items.call_count == 1

    # We expect manipulation to not revalidate immediately without strict
    foo.a.append("foo")
    foo.a.append("bar")
    assert validate_items.call_count == 1

    # We expect accessing data elements to cause a revalidate, but only the first time
    print(foo.a[0])
    assert validate_items.call_count == 2

    print(foo.a)
    assert foo.a == ["foo", "bar"]
    assert validate_items.call_count == 2


def test_strict_validation(mocker):
    """ Validate that when specified as strict, validation still occurs on every change"""
    schema = {
        "title": "Example Schema",
        "type": "object",
        "properties": {
            "a": {"type": "array", "items": {"type": "string"}, "default": []}
        },
    }

    ob = pjs.ObjectBuilder(schema)
    ns1 = ob.build_classes(strict=True)
    validator = ns1.ExampleSchema.a.info["validator"]([])
    validate_items = mocker.patch.object(
        validator, "validate_items", side_effect=validator.validate_items
    )
    validate = mocker.patch.object(
        validator, "validate", side_effect=validator.validate
    )
    mocker.patch.dict(
        ns1.ExampleSchema.a.info,
        {"validator": mocker.MagicMock(return_value=validator)},
    )

    foo = ns1.ExampleSchema()
    # We expect validation to be called on creation
    assert validate_items.call_count == 1

    # We expect manipulation to revalidate immediately with strict
    foo.a.append("foo")
    foo.a.append("bar")
    assert validate_items.call_count == 3

    # We expect accessing data elements to not revalidate because strict would have revalidated on load
    print(foo.a[0])
    print(foo.a)
    assert foo.a == ["foo", "bar"]



import markdown
from markdown.extensions import Extension
from markdown.preprocessors import Preprocessor
import re
import json


def extract_code_blocks(filename):
    with open(filename) as fin:
        doc = fin.read().split("\n")

    M = markdown.Markdown(extensions=[SpecialFencedCodeExtension()])

    preprocessors = M.preprocessors
    tree_processors = M.treeprocessors

    try:
        version_info = markdown.__version_info__
    except AttributeError:
        version_info = markdown.version_info

    # Markdown 3.* stores the processors in a class that can be iterated directly.
    # Markdown 2.* stores them in a dict, so we have to pull out the values.
    if version_info[0] == 2:
        # Note: `markdown.version_info` will be deprecated in favor of
        # `markdown.__version_info__` in later versions of Markdown.
        preprocessors = preprocessors.values()
        tree_processors = tree_processors.values()

    for prep in preprocessors:
        doc = prep.run(doc)

    root = M.parser.parseDocument(doc).getroot()

    for treeproc in tree_processors:
        newRoot = treeproc.run(root)
        if newRoot is not None:
            root = newRoot

    return SpecialFencePreprocessor.EXAMPLES


class SpecialFencedCodeExtension(Extension):
    def extendMarkdown(self, md, md_globals=None):
        """ Add FencedBlockPreprocessor to the Markdown instance. """
        md.registerExtension(self)

        if markdown.version_info[0] >= 3:
            md.preprocessors.register(
                SpecialFencePreprocessor(md), "fenced_code_block", 10
            )
        else:
            md.preprocessors.add(
                "fenced_code_block",
                SpecialFencePreprocessor(md),
                ">normalize_whitespace",
            )


class SpecialFencePreprocessor(Preprocessor):
    EXAMPLES = {}
    FENCED_BLOCK_RE = re.compile(
        r"""
(?P<fence>^(?:~{3,}|`{3,}))[ ]*         # Opening ``` or ~~~
(\{?\.?(?P<lang>[a-zA-Z0-9_+-]*))?[ ]*  # Optional {, and lang
# Optional highlight lines, single- or double-quote-delimited
(hl_lines=(?P<quot>"|')(?P<hl_lines>.*?)(?P=quot))?[ ]*
}?[ ]*\n                                # Optional closing }
(?P<code>.*?)(?<=\n)
(?P=fence)[ ]*$""",
        re.MULTILINE | re.DOTALL | re.VERBOSE,
    )

    def __init__(self, md):
        super(SpecialFencePreprocessor, self).__init__(md)

        self.checked_for_codehilite = False
        self.codehilite_conf = {}

    def run(self, lines):

        text = "\n".join(lines)

        while True:
            m = self.FENCED_BLOCK_RE.search(text)
            if m:
                if m.group("lang"):
                    lang = m.group("lang")
                    example = m.group("code")
                    try:
                        self.EXAMPLES[lang].append(example)
                    except KeyError:
                        self.EXAMPLES[lang] = [example]

                text = "%s\n%s" % (text[: m.start()], text[m.end() :])
            else:
                break
        return text.split("\n")


if __name__ == "__main__":
    print(extract_code_blocks())



import pytest

import json
import pkg_resources
import os
import python_jsonschema_objects as pjs
import python_jsonschema_objects.markdown_support


@pytest.fixture
def markdown_examples():
    md = pkg_resources.resource_filename(
        "python_jsonschema_objects.examples", "README.md"
    )
    examples = python_jsonschema_objects.markdown_support.extract_code_blocks(md)
    examples = {json.loads(v)["title"]: json.loads(v) for v in examples["schema"]}
    return examples


@pytest.fixture(autouse=True)
def inject_examples(doctest_namespace, markdown_examples):
    doctest_namespace["examples"] = markdown_examples


@pytest.fixture
def Person(markdown_examples):
    builder = pjs.ObjectBuilder(
        markdown_examples["Example Schema"], resolved=markdown_examples
    )
    assert builder
    return builder.classes["ExampleSchema"]



import pytest

from jsonschema import validate
import python_jsonschema_objects as pjo
import json


@pytest.fixture
def bad_schema_49():
    return {
        "title": "example",
        "type": "array",
        "items": {
            "oneOf": [
                {
                    "type": "object",
                    "properties": {"a": {"type": "string"}},
                    "required": ["a"],
                },
                {
                    "type": "object",
                    "properties": {"b": {"type": "string"}},
                    "required": ["b"],
                },
            ]
        },
    }


@pytest.fixture
def instance():
    return [{"a": ""}, {"b": ""}]


def test_is_valid_jsonschema(bad_schema_49, instance):
    validate(instance, bad_schema_49)


def test_regression_49(bad_schema_49, instance):
    builder = pjo.ObjectBuilder(bad_schema_49)
    ns = builder.build_classes()

    ns.Example.from_json(json.dumps(instance))



import codecs
import copy
import json
import logging
import os.path
import warnings

import inflection
import jsonschema
import six
from jsonschema import Draft4Validator

import python_jsonschema_objects.classbuilder as classbuilder
import python_jsonschema_objects.markdown_support
import python_jsonschema_objects.util
from python_jsonschema_objects.validators import ValidationError


logger = logging.getLogger(__name__)

__all__ = ["ObjectBuilder", "markdown_support", "ValidationError"]

FILE = __file__

SUPPORTED_VERSIONS = (
    "http://json-schema.org/draft-03/schema#",
    "http://json-schema.org/draft-04/schema#",
)


class ObjectBuilder(object):
    def __init__(self, schema_uri, resolved={}, resolver=None, validatorClass=None):
        self.mem_resolved = resolved

        if isinstance(schema_uri, six.string_types):
            uri = os.path.normpath(schema_uri)
            self.basedir = os.path.dirname(uri)
            with codecs.open(uri, "r", "utf-8") as fin:
                self.schema = json.loads(fin.read())
        else:
            self.schema = schema_uri
            uri = os.path.normpath(FILE)
            self.basedir = os.path.dirname(uri)

        if (
            "$schema" in self.schema
            and self.schema["$schema"] not in SUPPORTED_VERSIONS
        ):
            warnings.warn(
                "Schema version {} not recognized. Some "
                "keywords and features may not be supported.".format(
                    self.schema["$schema"]
                )
            )

        self.resolver = resolver or jsonschema.RefResolver.from_schema(self.schema)
        self.resolver.handlers.update(
            {"file": self.relative_file_resolver, "memory": self.memory_resolver}
        )

        validatorClass = validatorClass or Draft4Validator
        meta_validator = validatorClass(validatorClass.META_SCHEMA)
        meta_validator.validate(self.schema)
        self.validator = validatorClass(self.schema, resolver=self.resolver)

        self._classes = None
        self._resolved = None

    @property
    def schema(self):
        try:
            return copy.deepcopy(self._schema)
        except AttributeError:
            raise ValidationError("No schema provided")

    @schema.setter
    def schema(self, val):
        setattr(self, "_schema", val)

    @property
    def classes(self):
        if self._classes is None:
            self._classes = self.build_classes()
        return self._classes

    def get_class(self, uri):
        if self._resolved is None:
            self._classes = self.build_classes()
        return self._resolved.get(uri, None)

    def memory_resolver(self, uri):
        return self.mem_resolved[uri[7:]]

    def relative_file_resolver(self, uri):
        path = os.path.join(self.basedir, uri[8:])
        with codecs.open(path, "r", "utf-8") as fin:
            result = json.loads(fin.read())
        return result

    def validate(self, obj):
        try:
            return self.validator.validate(obj)
        except jsonschema.ValidationError as e:
            raise ValidationError(e)

    def build_classes(self, strict=False, named_only=False, standardize_names=True):
        """
        Build all of the classes named in the JSONSchema.

        Class names will be transformed using inflection by default, so names
        with spaces in the schema will be camelcased, while names without
        spaces will have internal capitalization dropped. Thus "Home Address"
        becomes "HomeAddress", while "HomeAddress" becomes "Homeaddress" To
        disable this behavior, pass standardize_names=False, but be aware
        that accessing names with spaces from the namespace can be
        problematic.

        Args:
            strict: (bool) use this to validate required fields while creating the class
            named_only: (bool) If true, only properties with an actual title attribute will
                be included in the resulting namespace (although all will be generated).
            standardize_names: (bool) If true (the default), class names will be tranformed
                by camel casing

        Returns:
            A namespace containing all the generated classes

        """
        kw = {"strict": strict}
        builder = classbuilder.ClassBuilder(self.resolver)
        for nm, defn in six.iteritems(self.schema.get("definitions", {})):
            uri = python_jsonschema_objects.util.resolve_ref_uri(
                self.resolver.resolution_scope, "#/definitions/" + nm
            )
            builder.construct(uri, defn, **kw)

        if standardize_names:
            name_transform = lambda t: inflection.camelize(
                inflection.parameterize(six.text_type(t), "_")
            )
        else:
            name_transform = lambda t: t

        nm = self.schema["title"] if "title" in self.schema else self.schema["$id"]
        nm = inflection.parameterize(six.text_type(nm), "_")

        builder.construct(nm, self.schema, **kw)
        self._resolved = builder.resolved

        classes = {}
        for uri, klass in six.iteritems(builder.resolved):
            title = getattr(klass, "__title__", None)
            if title is not None:
                classes[name_transform(title)] = klass
            elif not named_only:
                classes[name_transform(uri.split("/")[-1])] = klass

        return python_jsonschema_objects.util.Namespace.from_mapping(classes)


if __name__ == "__main__":

    validator = ObjectBuilder("../../protocol/json/schema.json")

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



import pytest

from jsonschema import validate
import python_jsonschema_objects as pjo
import json
import logging


@pytest.fixture
def nested_arrays():
    return {
        "title": "example",
        "properties": {
            "foo": {
                "type": "array",
                "items": {
                    "type": "array",
                    "items": [{"type": "number"}, {"type": "number"}],
                },
            }
        },
    }


@pytest.fixture
def instance():
    return {"foo": [[42, 44]]}


def test_validates(nested_arrays, instance):
    validate(instance, nested_arrays)


def test_nested_array_regression(nested_arrays, instance):
    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger().setLevel(logging.DEBUG)
    builder = pjo.ObjectBuilder(nested_arrays)
    ns = builder.build_classes()

    q = ns.Example.from_json(json.dumps(instance))

    assert q.serialize() == '{"foo": [[42, 44]]}'

    assert q.as_dict() == {"foo": [[42, 44]]}


@pytest.fixture
def complex_schema():
    return json.loads(
        r'{"definitions": {"pnu_info": {"required": ["unit_name", "unit_type", "version", "system_time"], "type": "object", "properties": {"unit_type": {"enum": ["Other", "Backpack"], "type": "string"}, "unit_name": {"type": "string"}, "system_time": {"type": "string"}, "version": {"type": "string"}, "recording_state": {"type": "string"}}}, "error": {"additionalProperties": true, "required": ["message"], "type": "object", "properties": {"message": {"type": "string"}}}, "ptu_location": {"required": ["ptu_id", "latitude", "longitude"], "type": "object", "properties": {"latitude": {"type": "number"}, "ptu_id": {"type": "string"}, "longitude": {"type": "number"}, "orientation": {"minimum": 0, "type": "number", "description": "The orientation of this PTU (in degrees). 360 means *unknown*", "maximum": 360}}}, "geopath": {"items": {"required": ["lat", "lng"], "type": "object", "properties": {"lat": {"type": "number"}, "lng": {"type": "number"}}}, "type": "array", "description": "A path described by an ordered\\nlist of lat/long coordinates\\n"}}, "required": ["status", "boundary", "members", "name"], "type": "object", "properties": {"status": {"enum": ["pending", "active", "completed"]}, "boundary": {"$ref": "#/definitions/geopath"}, "name": {"type": "string"}, "members": {"minItems": 1, "items": {"type": "string"}, "type": "array"}}, "title": "mission"}'
    )


def test_array_wrapper(complex_schema):
    instance = {
        "scenario_config": {"location_master": "MOCK"},
        "status": "pending",
        "boundary": [
            {"lat": 38.8821, "lng": -77.11461},
            {"lat": 38.882403, "lng": -77.107867},
            {"lat": 38.876293, "lng": -77.1083},
            {"lat": 38.880834, "lng": -77.115043},
        ],
        "name": "Test1",
        "members": ["Frobnaz", "MOCK"],
    }

    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger().setLevel(logging.DEBUG)
    builder = pjo.ObjectBuilder(complex_schema)
    ns = builder.build_classes()
    m = ns.Mission(**instance)
    m.validate()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# <python_jsonschema_objects - An object wrapper for JSON Schema definitions>
# Copyright (C) <2014-2016>  Chris Wacek <cwacek@gmail.com

import ast
import os
import re
import sys
from setuptools import setup, find_packages

import versioneer


if __name__ == "__main__":
    if "register" in sys.argv or "upload" in sys.argv:
        import register

        try:
            long_description = register.markdown_to_rst("README.md")
            if len(long_description) < 1:
                raise Exception("Failed to convert README.md")
        except Exception as e:
            sys.stderr.write("Error: {0}\n".format(e))
            sys.exit(1)

    else:
        long_description = ""

    setup(
        name="python_jsonschema_objects",
        version=versioneer.get_version(),
        description="An object wrapper for JSON Schema definitions",
        author="Chris Wacek",
        long_description=long_description,
        license="MIT",
        author_email="cwacek@gmail.com",
        packages=find_packages(),
        include_package_data=True,
        package_data={"python_jsonschema_objects.examples": ["README.md"]},
        zip_safe=False,
        url="http://python-jsonschema-objects.readthedocs.org/",
        setup_requires=["setuptools>=18.0.0"],
        install_requires=[
            "inflection>=0.2",
            "Markdown>=2.4",
            "jsonschema>=2.3",
            "six>=1.5.2",
        ],
        cmdclass=versioneer.get_cmdclass(),
        classifiers=[
            "Programming Language :: Python :: 2",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Intended Audience :: Developers",
            "Development Status :: 4 - Beta",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
    )



import pytest

import python_jsonschema_objects as pjo


@pytest.fixture
def arrayClass():
    schema = {
        "title": "ArrayVal",
        "type": "object",
        "properties": {
            "min": {
                "type": "array",
                "items": {"type": "string"},
                "default": [],
                "minItems": 1,
            },
            "max": {
                "type": "array",
                "items": {"type": "string"},
                "default": [],
                "maxItems": 1,
            },
            "both": {
                "type": "array",
                "items": {"type": "string"},
                "default": [],
                "maxItems": 2,
                "minItems": 1,
            },
            "unique": {
                "type": "array",
                "items": {"type": "string"},
                "default": [],
                "uniqueItems": True,
            },
            "reffed": {
                "type": "array",
                "items": {"$ref": "#/definitions/myref"},
                "minItems": 1,
            },
        },
        "definitions": {"myref": {"type": "string"}},
    }

    ns = pjo.ObjectBuilder(schema).build_classes()
    return ns["Arrayval"](min=["1"], both=["1"])


def test_validators_work_with_reference(arrayClass):
    arrayClass.reffed = ["foo"]

    with pytest.raises(pjo.ValidationError):
        arrayClass.reffed = []


def test_array_length_validates(markdown_examples):

    builder = pjo.ObjectBuilder(
        markdown_examples["Example Schema"], resolved=markdown_examples
    )
    ns = builder.build_classes()

    with pytest.raises(pjo.ValidationError):
        ns.ExampleSchema(
            firstName="Fred",
            lastName="Huckstable",
            dogs=["Fido", "Spot", "Jasper", "Lady", "Tramp"],
        )


def test_minitems(arrayClass):

    arrayClass.min = ["1"]
    arrayClass.min.append("2")

    with pytest.raises(pjo.ValidationError):
        arrayClass.min = []


def test_maxitems(arrayClass):

    arrayClass.max = []
    arrayClass.max.append("2")

    assert arrayClass.max == ["2"]

    with pytest.raises(pjo.ValidationError):
        arrayClass.max.append("3")
        # You have to explicitly validate with append
        arrayClass.validate()

    with pytest.raises(pjo.ValidationError):
        arrayClass.max = ["45", "42"]


def test_unique(arrayClass):

    arrayClass.unique = ["hi", "there"]
    with pytest.raises(pjo.ValidationError):
        arrayClass.unique.append("hi")
        # You have to explicitly validate with append
        arrayClass.validate()

    with pytest.raises(pjo.ValidationError):
        arrayClass.unique = ["Fred", "Fred"]



import pandoc
import os


def markdown_to_rst(src):
    pandoc.core.PANDOC_PATH = "/usr/local/bin/pandoc"
    if not os.path.exists(pandoc.core.PANDOC_PATH):
        raise Exception("Pandoc not available")

    doc = pandoc.Document()
    doc.markdown = open("README.md").read()
    return doc.rst



import pytest
import python_jsonschema_objects as pjs
import collections


@pytest.fixture
def schema():
    return {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "title": "Test",
        "definitions": {
            "MyEnum1": {"type": "string", "enum": ["E_A", "E_B"]},
            "MyEnum2": {"type": "string", "enum": ["F_A", "F_B", "F_C", "F_D"]},
            "MyInt": {
                "default": "0",
                "type": "integer",
                "minimum": 0,
                "maximum": 4294967295,
            },
            "MyObj1": {
                "type": "object",
                "properties": {
                    "e1": {"$ref": "#/definitions/MyEnum1"},
                    "e2": {"$ref": "#/definitions/MyEnum2"},
                    "i1": {"$ref": "#/definitions/MyInt"},
                },
                "required": ["e1", "e2", "i1"],
            },
            "MyArray": {
                "type": "array",
                "items": {"$ref": "#/definitions/MyObj1"},
                "minItems": 0,
                "uniqueItems": True,
            },
            "MyMsg1": {
                "type": "object",
                "properties": {"a1": {"$ref": "#/definitions/MyArray"}},
            },
            "MyMsg2": {"type": "object", "properties": {"s1": {"type": "string"}}},
        },
        "type": "object",
        "oneOf": [{"$ref": "#/definitions/MyMsg1"}, {"$ref": "#/definitions/MyMsg2"}],
    }


def test_regression_126(schema):
    builder = pjs.ObjectBuilder(schema)
    ns = builder.build_classes(standardize_names=False)

    Obj1 = ns.MyObj1
    Array1 = ns.MyArray
    Msg1 = ns.MyMsg1
    o1 = Obj1(e1="E_A", e2="F_C", i1=2600)
    o2 = Obj1(e1="E_B", e2="F_D", i1=2500)
    objs = Array1([o1, o2])
    msg = Msg1(a1=objs)

    print(msg.serialize())



import json
import pytest
import python_jsonschema_objects as pjo

schema = """
{
    "title":"whatever",
    "properties": {
        "test": {"type": "number"}
    }
}
"""


@pytest.fixture
def schema_json():
    return json.loads(schema)


def test_literals_support_comparisons(schema_json):
    builder = pjo.ObjectBuilder(schema_json)
    ns = builder.build_classes()

    thing1 = ns.Whatever()
    thing2 = ns.Whatever()

    thing1.test = 10
    thing2.test = 12

    assert thing1.test < thing2.test
    assert thing1.test == thing1.test

    thing2.test = 10

    assert thing1.test <= thing2.test
    assert thing1.test == thing2.test



import pytest
import python_jsonschema_objects as pjs
import jsonschema

import json
import copy
import pprint

pp = pprint.PrettyPrinter(indent=2)

lpid_class_schema = {
    "$id": "https://example.com/schema-repository/lpid",
    "$schema": "http://json-schema.org/draft-04/schema#",
    "type": "object",
    "title": "Logical Process ID",
    "properties": {"lpid_str": {"type": "string"}},
    "required": ["lpid_str"],
    "additionalProperties": True,
}


class A(object):
    a_str = None

    def __init__(self, a_str):
        self.a_str = a_str


def test_regression_185_deepcopy():
    # Class building and instantiation work great and round-trip.
    lpid_class_builder = pjs.ObjectBuilder(lpid_class_schema)
    LPID = lpid_class_builder.classes.LogicalProcessId
    lpid = LPID(lpid_str="foobaz")
    instance = json.loads(lpid.serialize())
    jsonschema.validate(lpid_class_schema, instance)

    # Deep-copying seems to work with instances of normal popos.
    a = A(a_str="barqux")
    assert a
    a_2 = copy.deepcopy(a)
    assert a_2

    # But deep-copying with objects of generated classes doesn't terminate.
    lpid_2 = copy.deepcopy(lpid)
    assert repr(lpid_2) == repr(lpid)



import pytest

import warnings
import json
import six
import jsonschema
import python_jsonschema_objects as pjs

import logging

logging.basicConfig(level=logging.DEBUG)


@pytest.mark.parametrize(
    "version, warn",
    [
        ("http://json-schema.org/schema#", True),
        ("http://json-schema.org/draft-03/schema#", False),
        ("http://json-schema.org/draft-04/schema#", False),
        ("http://json-schema.org/draft-06/schema#", True),
        ("http://json-schema.org/draft-07/schema#", True),
    ],
)
def test_warnings_on_schema_version(version, warn):
    schema = {"$schema": version, "$id": "test", "type": "object", "properties": {}}

    with warnings.catch_warnings(record=True) as w:
        pjs.ObjectBuilder(schema)

        if warn:
            assert len(w) == 1
            assert "Schema version %s not recognized" % version in str(w[-1].message)
        else:
            assert len(w) == 0, w[-1].message


def test_schema_validation():
    """Test that the ObjectBuilder validates the schema itself."""
    schema = {
        "$schema": "http://json-schema.org/schema#",
        "$id": "test",
        "type": "object",
        "properties": {
            "name": "string",  #  <-- this is invalid
            "email": {"oneOf": [{"type": "string"}, {"type": "integer"}]},
        },
        "required": ["email"],
    }
    with pytest.raises(jsonschema.ValidationError):
        pjs.ObjectBuilder(schema)


def test_regression_9():
    schema = {
        "$schema": "http://json-schema.org/schema#",
        "$id": "test",
        "type": "object",
        "properties": {
            "name": {"type": "string"},
            "email": {"oneOf": [{"type": "string"}, {"type": "integer"}]},
        },
        "required": ["email"],
    }
    builder = pjs.ObjectBuilder(schema)
    builder.build_classes()


def test_build_classes_is_idempotent():
    schema = {
        "$schema": "http://json-schema.org/schema#",
        "title": "test",
        "type": "object",
        "properties": {
            "name": {"$ref": "#/definitions/foo"},
            "email": {"oneOf": [{"type": "string"}, {"type": "integer"}]},
        },
        "required": ["email"],
        "definitions": {
            "reffed": {"type": "string"},
            "foo": {"type": "array", "items": {"$ref": "#/definitions/reffed"}},
        },
    }
    builder = pjs.ObjectBuilder(schema)
    x = builder.build_classes()
    builder.build_classes()


def test_underscore_properties():
    schema = {
        "$schema": "http://json-schema.org/schema#",
        "title": "AggregateQuery",
        "type": "object",
        "properties": {"group": {"type": "object", "properties": {}}},
    }

    builder = pjs.ObjectBuilder(schema)
    ns = builder.build_classes()
    my_obj_type = ns.Aggregatequery
    request_object = my_obj_type(
        group={
            "_id": {"foo_id": "$foo_id", "foo_type": "$foo_type"},
            "foo": {"$sum": 1},
        }
    )

    assert request_object.group._id == {"foo_id": "$foo_id", "foo_type": "$foo_type"}


def test_array_regressions():
    schema = {
        "$schema": "http://json-schema.org/schema#",
        "$id": "test",
        "type": "object",
        "properties": {
            "name": {"type": "string"},
            "email_aliases": {
                "type": "object",
                "additionalProperties": {
                    "type": "array",
                    "items": {"$ref": "#/definitions/foo"},
                },
            },
        },
        "definitions": {"foo": {"type": "string"}},
    }
    builder = pjs.ObjectBuilder(schema)

    ns = builder.build_classes()

    x = ns.Test.from_json(
        """{"email_aliases": {
            "Freddie": ["james", "bond"]
            }}"""
    )
    x.validate()

    y = ns.Test(email_aliases={"Freddie": ["james", "bond"]})
    y.validate()


def test_arrays_can_have_reffed_items_of_mixed_type():
    schema = {
        "$schema": "http://json-schema.org/schema#",
        "$id": "test",
        "type": "object",
        "properties": {
            "list": {
                "type": "array",
                "items": {
                    "oneOf": [
                        {"$ref": "#/definitions/foo"},
                        {
                            "type": "object",
                            "properties": {"bar": {"type": "string"}},
                            "required": ["bar"],
                        },
                    ]
                },
            }
        },
        "definitions": {"foo": {"type": "string"}},
    }
    builder = pjs.ObjectBuilder(schema)
    ns = builder.build_classes()

    ns.Test(list=["foo", "bar"])
    ns.Test(list=[{"bar": "nice"}, "bar"])
    with pytest.raises(pjs.ValidationError):
        ns.Test(list=[100])


def test_regression_39():
    builder = pjs.ObjectBuilder("test/thing-two.json")
    ns = builder.build_classes()

    for thing in ("BarMessage", "BarGroup", "Bar", "Header"):
        assert thing in ns

    x = ns.BarMessage(
        id="message_id", title="my bar group", bars=[{"name": "Freddies Half Shell"}]
    )

    x.validate()

    # Now an invalid one
    with pytest.raises(pjs.ValidationError):
        ns.BarMessage(
            id="message_id",
            title="my bar group",
            bars=[{"Address": "I should have a name"}],
        )


def test_loads_markdown_schema_extraction(markdown_examples):
    assert "Other" in markdown_examples


def test_object_builder_loads_memory_references(markdown_examples):
    builder = pjs.ObjectBuilder(markdown_examples["Other"], resolved=markdown_examples)
    assert builder

    with pytest.raises(pjs.ValidationError):
        builder.validate({"MyAddress": 1234})

    builder.validate({"MyAddress": "1234"})


def test_object_builder_reads_all_definitions(markdown_examples):
    for nm, ex in six.iteritems(markdown_examples):
        builder = pjs.ObjectBuilder(ex, resolved=markdown_examples)
        assert builder


@pytest.fixture
def oneOf(markdown_examples):
    builder = pjs.ObjectBuilder(markdown_examples["OneOf"], resolved=markdown_examples)
    return builder.classes["Oneof"]


@pytest.mark.parametrize(
    "json_object", ['{"MyData": "an address"}', '{"MyData": "1234"}']
)
def test_oneOf_validates_against_any_valid(oneOf, json_object):

    oneOf.from_json(json_object)


def test_oneOf_fails_against_non_matching(oneOf):
    with pytest.raises(pjs.ValidationError):
        oneOf.from_json('{"MyData": 1234.234}')


@pytest.fixture
def oneOfBare(markdown_examples):
    builder = pjs.ObjectBuilder(
        markdown_examples["OneOfBare"], resolved=markdown_examples
    )
    return builder.classes["Oneofbare"]


@pytest.mark.parametrize(
    "json_object",
    ['{"MyAddress": "an address"}', '{"firstName": "John", "lastName": "Winnebago"}'],
)
def test_oneOfBare_validates_against_any_valid(oneOfBare, json_object):

    oneOfBare.from_json(json_object)


def test_oneOfBare_fails_against_non_matching(oneOfBare):
    with pytest.raises(pjs.ValidationError):
        oneOfBare.from_json('{"MyData": 1234.234}')


@pytest.fixture
def Other(markdown_examples):
    builder = pjs.ObjectBuilder(markdown_examples["Other"], resolved=markdown_examples)
    assert builder
    return builder.classes["Other"]


def test_additional_props_allowed_by_default(Person):
    person = Person()
    person.randomAttribute = 4
    assert int(person.randomAttribute) == 4


def test_additional_props_permitted_explicitly(markdown_examples):
    builder = pjs.ObjectBuilder(
        markdown_examples["AddlPropsAllowed"], resolved=markdown_examples
    )
    assert builder

    test = builder.classes["Addlpropsallowed"]()
    test.randomAttribute = 40
    assert int(test.randomAttribute) == 40


def test_still_raises_when_accessing_undefined_attrs(Person):

    person = Person()
    person.firstName = "James"

    # If the attribute doesn't exist, we expect an AttributeError
    with pytest.raises(AttributeError):
        print(person.randomFoo)

    # If the attribute is literal-esq, isn't set but isn't required, accessing it should be fine
    assert person.gender == None

    # If the attribute is an object, isn't set, but isn't required accessing it should throw an exception
    with pytest.raises(AttributeError) as e:
        print(person.address.street)
        assert "'NoneType' object has no attribute 'street'" in e


def test_permits_deletion_of_additional_properties(Person):
    person = Person()
    person.randomthing = 44
    assert int(person.randomthing) == 44

    del person["randomthing"]

    with pytest.raises(AttributeError):
        assert person.randomthing is None


def test_additional_props_disallowed_explicitly(Other):

    other = Other()
    with pytest.raises(pjs.ValidationError):
        other.randomAttribute = 4


def test_objects_can_be_empty(Person):
    assert Person()


def test_object_equality_should_compare_data(Person):
    person = Person(firstName="john")
    person2 = Person(firstName="john")

    assert person == person2

    person2.lastName = "Wayne"
    assert person != person2


def test_object_allows_attributes_in_oncstructor(Person):
    person = Person(firstName="James", lastName="Bond", age=35)

    assert person
    assert str(person.firstName) == "James"
    assert str(person.lastName) == "Bond"
    assert int(person.age) == 35


def test_object_validates_on_json_decode(Person):
    with pytest.raises(pjs.ValidationError):
        Person.from_json('{"firstName": "James"}')


def test_object_validates_enumerations(Person):
    person = Person()

    with pytest.raises(pjs.ValidationError):
        person.gender = "robot"

    person.gender = "male"
    person.gender = "female"


def test_validation_of_mixed_type_enums(Person):
    person = Person()

    person.deceased = "yes"
    person.deceased = "no"
    person.deceased = 1

    with pytest.raises(pjs.ValidationError):
        person.deceased = "robot"

    with pytest.raises(pjs.ValidationError):
        person.deceased = 2

    with pytest.raises(pjs.ValidationError):
        person.deceased = 2.3


def test_objects_allow_non_required_attrs_to_be_missing(Person):
    person = Person(firstName="James", lastName="Bond")

    assert person
    assert str(person.firstName) == "James"
    assert str(person.lastName) == "Bond"


def test_objects_require_required_attrs_on_validate(Person):
    person = Person(firstName="James")

    assert person

    with pytest.raises(pjs.ValidationError):
        person.validate()


@pytest.fixture(scope="function")
def person_object(Person):
    return Person(firstName="James")


def test_attribute_access_via_dict(person_object):
    name = person_object["firstName"]
    assert str(name) == "James"


def test_attribute_set_via_dict(person_object):
    person_object["firstName"] = "John"

    name = person_object["firstName"]
    assert str(name) == "John"


def test_keys_can_be_other_pjo_objects(Person, person_object):
    person_object.lastName = "Smith"

    jane = Person(firstName="Jane", lastName=person_object.lastName)

    assert jane.lastName == person_object.lastName

    # We don't want the names to be the same literal object though.
    # Changing one after assignment should not change the other
    jane.lastName = "Borges"
    assert jane.lastName != person_object.lastName


def test_numeric_attribute_validation(Person):
    with pytest.raises(pjs.ValidationError):
        Person(firstName="James", lastName="Bond", age=-10)

    p = Person(firstName="James", lastName="Bond")
    with pytest.raises(pjs.ValidationError):
        p.age = -1


def test_objects_validate_prior_to_serializing(Person):
    p = Person(firstName="James", lastName="Bond")

    p._properties["age"] = -1

    with pytest.raises(pjs.ValidationError):
        p.serialize()


def test_serializing_removes_null_objects(Person):
    person = Person(firstName="James", lastName="Bond")

    json_str = person.serialize()
    assert "age" not in json.loads(json_str)


def test_lists_get_serialized_correctly(Person):
    person = Person(firstName="James", lastName="Bond", dogs=["Lassie", "Bobo"])

    json_str = person.serialize()
    assert json.loads(json_str) == {
        "firstName": "James",
        "lastName": "Bond",
        "dogs": ["Lassie", "Bobo"],
    }


@pytest.mark.parametrize(
    "pdict",
    [
        dict(firstName="James", lastName="Bond", dogs=["Lassie", "Bobo"]),
        dict(
            firstName="James",
            lastName="Bond",
            address={
                "street": "10 Example Street",
                "city": "Springfield",
                "state": "USA",
            },
            dogs=["Lassie", "Bobo"],
        ),
    ],
)
def test_dictionary_transformation(Person, pdict):

    person = Person(**pdict)

    assert person.as_dict() == pdict


def test_strict_mode():
    schema = {
        "$schema": "http://json-schema.org/schema#",
        "type": "object",
        "properties": {"firstName": {"type": "string"}, "lastName": {"type": "string"}},
        "$id": "test",
        "title": "Name Data",
        "required": ["firstName"],
    }
    builder = pjs.ObjectBuilder(schema)
    ns = builder.build_classes()  # by defualt strict = False
    NameData = ns.NameData
    # no strict flag - so should pass even no firstName
    NameData(lastName="hello")
    with pytest.raises(pjs.ValidationError):
        ns = builder.build_classes(strict=True)
        NameData = ns.NameData
        NameData(lastName="hello")


def test_boolean_in_child_object():
    schema = {
        "$schema": "http://json-schema.org/schema#",
        "$id": "test",
        "type": "object",
        "properties": {"data": {"type": "object", "additionalProperties": True}},
    }
    builder = pjs.ObjectBuilder(schema)
    ns = builder.build_classes()

    ns.Test(data={"my_bool": True})


@pytest.mark.parametrize(
    "default",
    [
        '{"type": "boolean", "default": false}',
        '{"type": "string", "default": "Hello"}',
        '{"type": "integer", "default": 500}',
    ],
)
def test_default_values(default):
    default = json.loads(default)
    schema = {
        "$schema": "http://json-schema.org/schema#",
        "$id": "test",
        "type": "object",
        "properties": {"sample": default},
    }

    builder = pjs.ObjectBuilder(schema)
    ns = builder.build_classes()

    x = ns.Test()
    assert x.sample == default["default"]


def test_justareference_example(markdown_examples):

    builder = pjs.ObjectBuilder(
        markdown_examples["Just a Reference"], resolved=markdown_examples
    )
    ns = builder.build_classes()
    ns.JustAReference("Hello")



import pytest

import python_jsonschema_objects as pjs


def test_multiple_objects_are_defined(markdown_examples):
    builder = pjs.ObjectBuilder(
        markdown_examples["MultipleObjects"], resolved=markdown_examples
    )

    assert builder
    classes = builder.build_classes()
    assert "ErrorResponse" in classes
    assert "VersionGetResponse" in classes
    print(dir(classes))



import pytest
import python_jsonschema_objects as pjo


def test_simple_array_oneOf():
    basicSchemaDefn = {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "title": "Test",
        "properties": {
            "SimpleArrayOfNumberOrString": {"$ref": "#/definitions/simparray"}
        },
        "required": ["SimpleArrayOfNumberOrString"],
        "type": "object",
        "definitions": {
            "simparray": {
                "oneOf": [
                    {"type": "array", "items": {"type": "number"}},
                    {"type": "array", "items": {"type": "string"}},
                ]
            }
        },
    }

    builder = pjo.ObjectBuilder(basicSchemaDefn)

    ns = builder.build_classes()
    ns.Test().from_json('{"SimpleArrayOfNumberOrString" : [0, 1]}')
    ns.Test().from_json('{"SimpleArrayOfNumberOrString" : ["Hi", "There"]}')

    with pytest.raises(pjo.ValidationError):
        ns.Test().from_json('{"SimpleArrayOfNumberOrString" : ["Hi", 0]}')
